/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_BWB_BWBDEFAULTSTRATEGY_H_
#define CALCULATEPOLAR_SRC_BWB_BWBDEFAULTSTRATEGY_H_

#include "bwbCalculatePolarConfig.h"

#include <moduleBasics/runtimeIO.h>

#include <string>
#include <unordered_map>
#include <vector>
#include <map>
#include <memory>

#include "bwbDefaultData.h"
#include "../polar.h"
#include "../methods/liftingLinePolar.h"
#include "../methods/viscDragRaymer.h"
#include "../methods/waveDragMason.h"
#include "../methods/semiEmpiricalHighLiftAdaption.h"

class bwbDefaultStrategy {
 public:
    // geometry bools
    // containers for the aircraft geometry
    std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>> liftingSurfaces;    /**< vector which stores all lifting surface geometries (wing and HTP)*/
    geom2::MultisectionSurface<geom2::AirfoilSection> VTP;                              /**< the VTP geometry */
    // container for polar data
    struct PolarData {                                                                  /**< struct for handling of the polar data */
        std::unordered_map<size_t, tawDefaultPolar> cleanPolars;                        /**< clean polars indexed by MachID */
        std::unordered_map<std::string, tawDefaultPolar> highLiftPolars;                /**< high lift polars indexed by config */
    };
    PolarData currentPolarSet;                                                          /**< temporary polar set which gets overwritten for each new calculation run */
    std::unordered_map<size_t, PolarData> polarDataTrimMap;                             /**< Polar data ordered by stabilizer angle index */
    std::unordered_map<size_t, std::unordered_map<size_t, PolarData>> polarDataCGMap;   /**< Polar data ordered by center of gravity index (0: design, 1: forward, 2; aft) */
                                                                                                                // and by stabilizer angle index
    std::vector<PolarData> trimmedPolars;                                               /**< trimmed polars ordered by center of gravity index */
    std::vector<std::string> HLconfigs;
    std::string polarXMLFilePath;
    std::string designPolarXMLFilePath;
    std::string LILI_Version;
    void run();
    /** \brief Function initializes liftingLineForTAW with the required variables
     *  \param myLiftingLineForTAW the lifting line for TAW method with gets initialized
     */
    void initializeLiLiPolar(LiLiPolarInitialization *myLiLiPolarInit);
    /** \brief Function initializes viscDragRaymer with the required variables
     *  \param myViscDragRaymer the viscous drag raymer method with gets initialized
     */
    void initializeViscDragRaymer(viscDragRaymerInitialization *myViscDragInit);
    /** \brief Function initializes waveDragMason with the required variables
     *  \param myWaveDragMason the wave drag mason method with gets initialized
     */
    void initializeWaveDragMason(waveDragInitialization *myWaveDragInit);
    /** \brief Function with builds and saves the entire aircraft geometry
     */
    void createAircraftGeometry();
    /** \brief Function witch executes all aero calculation methods
     *  \param CoGPosition the 3D center of gravity coordinate as reference for the aero moments [m]
     */    
    void runAeroCalculations(const Point CoGPosition);
    /** \brief Function to copy the polar set from liftingLineForTAW to the strategyClass
     *  \param liftingLineForTAW the lifting line for TAW method, which generated the clean polar set
     */
    void getPolarDataFromLiftingLine(const liftingLinePolar& liftingLinePolar);
    /** \brief Function to calculate the viscous drag for the current polar set
     *  \param myViscDragRaymer the viscous drag raymer method to calculate the drag
     */
    void calculateViscDragRaymer(viscDragRaymer *myViscDragRaymer);
    /** \brief Function to calculate the wave drag for the current polar set
     *  \param myWaveDragMason the wave drag mason method to calculate the drag
     */
    void calculateWaveDragMason(waveDragMason *myWaveDragMason);

    void sumUpDragComponents();
    /** \brief Function copies the clean polars from the temporary current polar set to the permanent trimmed polar data map
     */
    void copyCleanPolars();
    void copyHighLiftPolars();
    /** \brief Function copies the high lift polars from the temporary current polar set to the permanent trimmed polar data map
     */
    void copyWingDevicesAndStabilizer(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption);
    void copyCleanPolarForHLCalculation(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, liftingLinePolar *myLiftingLinePolar);
    void calculateHighLiftPolars(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, const double CoG_X_Position);
    

    void writePolarXML();

    void writePolarData();

    void writeXMLOutputData();

    double getCLoptCruise();

    bwbDefaultStrategy(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<bwbCalculatePolarConfig>& config);
    virtual ~bwbDefaultStrategy() {}

 private:
    const std::shared_ptr<RuntimeIO>& rtIO_;                    /**< Pointer to the RuntimeIO */
    const std::shared_ptr<bwbDefaultData> data_;        /**< Pointer to the aircraft data */
    const std::shared_ptr<bwbCalculatePolarConfig>& config_;    /**< Pointer to the config data */

};


#endif  // CALCULATEPOLAR_SRC_BWB_BWBDEFAULTSTRATEGY_H_
