/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "HLFCData.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <vector>
#include <memory>
#include <utility>

using std::endl;

HLFCData::HLFCData(const node& acftXML)
    :
    bestCruiseCL(EndnodeReadOnly<double>("aerodynamics/lift_coefficients/C_LoptimumCruise")),
    initialCruiseMachNumber(EndnodeReadOnly<double>("top_level_aircraft_requirements/design_mission/initial_cruise_mach_number")),
    engineType(EndnodeReadOnly<std::string>("design_specification/propulsion/propulsor@0/type")),
    designCoGX(EndnodeReadOnly<double>("masses_cg_inertia/design_mass/mass_properties/center_of_gravity/x")),
    designCoGY(EndnodeReadOnly<double>("masses_cg_inertia/design_mass/mass_properties/center_of_gravity/y")),
    designCoGZ(EndnodeReadOnly<double>("masses_cg_inertia/design_mass/mass_properties/center_of_gravity/z")),
    forwardCoGX(EndnodeReadOnly<double>("masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x")),
    forwardCoGY(EndnodeReadOnly<double>("masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/y")),
    forwardCoGZ(EndnodeReadOnly<double>("masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/z")),
    aftCoGX(EndnodeReadOnly<double>("masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/x")),
    aftCoGY(EndnodeReadOnly<double>("masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/y")),
    aftCoGZ(EndnodeReadOnly<double>("masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/z")) {
    bestCruiseCL.read(acftXML);
    initialCruiseMachNumber.read(acftXML);
    engineType.read(acftXML);
    designCoGX.read(acftXML);
    designCoGY.read(acftXML);
    designCoGZ.read(acftXML);
    forwardCoGX.read(acftXML);
    forwardCoGY.read(acftXML);
    forwardCoGZ.read(acftXML);
    aftCoGX.read(acftXML);
    aftCoGY.read(acftXML);
    aftCoGZ.read(acftXML);
    designCoG.xCoordinate = designCoGX.value();
    designCoG.yCoordinate = designCoGY.value();
    designCoG.zCoordinate = designCoGZ.value();
    forwardCoG.xCoordinate = forwardCoGX.value();
    forwardCoG.yCoordinate = forwardCoGY.value();
    forwardCoG.zCoordinate = forwardCoGZ.value();
    aftCoG.xCoordinate = aftCoGX.value();
    aftCoG.yCoordinate = aftCoGY.value();
    aftCoG.zCoordinate = aftCoGZ.value();
    CoGPositions["design"] = designCoG;
    CoGPositions["forward"] = forwardCoG;
    CoGPositions["aft"] = aftCoG;
    numberDevices = acftXML.getVector(("wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device"), 3).size();
    // read in device names to later distinguish between control and high lift device
    for (size_t deviceID(0); deviceID < numberDevices; ++deviceID) {
        wingDevice aWingDevice;
        aWingDevice.type =
                    std::string(acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                    + num2Str(deviceID) + "/type/value"));
        aWingDevice.maxNegativeDeflection =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_negative_deflection/value");
        aWingDevice.maxPositiveDeflection =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_positive_deflection/value");
        aWingDevice.innerRelSpanPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/spanwise/value");
        aWingDevice.innerRelFwdChordPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/from/value");
        aWingDevice.innerRelAftChordPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/to/value");
        aWingDevice.outerRelSpanPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/spanwise/value");
        aWingDevice.outerRelFwdChordPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/from/value");
        aWingDevice.outerRelAftChordPosition =
                    acftXML.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/to/value");
        wingDevices.push_back(aWingDevice);
    }
}
