/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "trimInterpolation.h"

#include "generalMethods/auxFunctions.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>

#include <numeric>
#include <algorithm>
#include <vector>
#include <exception>
#include <iomanip>
#include <memory>
#include <string>

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::vector;

std::vector<trimInterpolation::interpolationPoint> trimInterpolation::interpolateTrim(const std::vector<std::vector<trimInterpolation::interpolationPoint>>& aPolarSet,
                                                                                        const bool allowExtrapolation) {
    thePolarSet = aPolarSet;
    extrapolationAllowed = allowExtrapolation;
    // trim off CL values of the Polar set
    this->cutCLRange();
    // find interpolation partners based on the CM values
    this->findInterpolationPartners();
    // interpolate
    this->interpolatePolarSet();
    return theTrimmedPolar;
}

void trimInterpolation::cutCLRange() {
    vector<double>minCL, maxCL;
    for (size_t polarID(0); polarID < thePolarSet.size(); ++polarID) {
        // check if CL is in ascending order and reverse if necessary
        if (thePolarSet.at(polarID).front().CL > thePolarSet.at(polarID).back().CL) {
            std::reverse(thePolarSet.at(polarID).begin(), thePolarSet.at(polarID).end());
        }
        // push back maximum and minimum CL values
        minCL.push_back(thePolarSet.at(polarID).front().CL);
        maxCL.push_back(thePolarSet.at(polarID).back().CL);
    }
    // determine the range CL range, so that no empty polars entrys are avaliable
    double totalMinCL = *max_element(minCL.begin(), minCL.end());
    double totalMaxCL = *min_element(maxCL.begin(), maxCL.end());
    // trim of the polars
    for (size_t polarID(0); polarID < thePolarSet.size(); ++polarID) {
        size_t minID(0), maxID(0);
        for (size_t pointID(0); pointID < thePolarSet.at(polarID).size(); ++pointID) {
            if (accuracyCheck(totalMinCL, thePolarSet.at(polarID).at(pointID).CL, ACCURACY_HIGH)) {
                minID = pointID;
            }
            if (accuracyCheck(totalMaxCL, thePolarSet.at(polarID).at(pointID).CL, ACCURACY_HIGH)) {
                maxID = pointID;
            }
        }
        // resize the vector
        thePolarSet.at(polarID).erase(thePolarSet.at(polarID).begin() + maxID, thePolarSet.at(polarID).end());
        thePolarSet.at(polarID).erase(thePolarSet.at(polarID).begin(), thePolarSet.at(polarID).begin() + minID);
    }
}

void trimInterpolation::findInterpolationPartners() {
    // check if polars in polar set have same size and same CL range
    size_t polarSize = thePolarSet.front().size();
    double minCL = thePolarSet.front().front().CL;
    double maxCL = thePolarSet.front().back().CL;
    for (size_t polarID(1); polarID < thePolarSet.size(); ++polarID) {
        if (polarSize != thePolarSet.at(polarID).size()) {
            stringstream errorMessage;
            errorMessage << "The Polarset has different sized polars in it, abort trim interpolation!";
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
        if ((minCL != thePolarSet.at(polarID).front().CL) || (maxCL != thePolarSet.at(polarID).back().CL)) {
            stringstream errorMessage;
            errorMessage << "The Polarset has different minimum or maximum CL values in its polars, abort trim interpolation!";
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    }
    // determine for each CL value, where in the polar set CM flips from negative to positive, or vice versa
    for (size_t pointID(0); pointID < thePolarSet.front().size(); ++pointID) {
        bool firstSignBit = std::signbit(thePolarSet.front().at(pointID).CM);
        bool foundInterpolationPartners = false;
        for (size_t polarID(1); polarID < thePolarSet.size(); ++polarID) {
            if (firstSignBit != std::signbit(thePolarSet.at(polarID).at(pointID).CM)) {
                interpolationPartners.push_back(std::make_pair(polarID - 1, polarID));
                foundInterpolationPartners = true;
                break;
            }
        }
        // check if extrapolation is allowed, if it is necessary
        if ((foundInterpolationPartners == false) &&(extrapolationAllowed == false)) {
            stringstream errorMessage;
            errorMessage << "Extrapolation is needed to trim the polar set, which is forbidden, abord trim interpolation!";
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
        // find extrapolation partners
        if (foundInterpolationPartners == false) {
            // determine wich CM at given polar point is closer to zero
            if (fabs(thePolarSet.front().at(pointID).CM) < fabs(thePolarSet.back().at(pointID).CM)) {
                interpolationPartners.push_back(std::make_pair(0, 1));
            } else {
                interpolationPartners.push_back(std::make_pair(thePolarSet.size() - 2, thePolarSet.size() - 1));
            }
        }
    }
}

void trimInterpolation::interpolatePolarSet() {
    // loop over complete polar
    for (size_t pointID(0); pointID < thePolarSet.front().size(); ++pointID) {
        interpolationPoint aPolarPoint;
        aPolarPoint.CL = thePolarSet.front().at(pointID).CL;
        aPolarPoint.CM = 0.0;
        // interpolate each value stored in the interpolation point vector
        for (size_t valueID(0); valueID < thePolarSet.front().front().values.size(); ++valueID) {
            aPolarPoint.values.push_back(linearInterpolation(0.0, thePolarSet.at(interpolationPartners.at(pointID).first).at(pointID).CM,
                                                                  thePolarSet.at(interpolationPartners.at(pointID).first).at(pointID).values.at(valueID),
                                                                  thePolarSet.at(interpolationPartners.at(pointID).second).at(pointID).CM,
                                                                  thePolarSet.at(interpolationPartners.at(pointID).second).at(pointID).values.at(valueID), true));
        }
        // interpolate the vector values stored in the interpolation point vector
        for (size_t vectorValueID(0); vectorValueID < thePolarSet.front().front().vectorValues.size(); ++vectorValueID) {
            aPolarPoint.vectorValues.push_back(vector<double>());
            for (size_t vectorID(0); vectorID < thePolarSet.front().front().vectorValues.at(vectorValueID).size(); ++vectorID) {
                aPolarPoint.vectorValues.back().push_back(linearInterpolation(0.0,
                                                        thePolarSet.at(interpolationPartners.at(pointID).first).at(pointID).CM,
                                                        thePolarSet.at(interpolationPartners.at(pointID).first).at(pointID).vectorValues.at(vectorValueID).at(vectorID),
                                                        thePolarSet.at(interpolationPartners.at(pointID).second).at(pointID).CM,
                                                        thePolarSet.at(interpolationPartners.at(pointID).second).at(pointID).vectorValues.at(vectorValueID).at(vectorID), true));
            }
        }
        theTrimmedPolar.push_back(aPolarPoint);
    }
}

trimInterpolation::trimInterpolation() {
    // ctor
}
