/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_METHODS_VISCDRAGRAYMER_H_
#define CALCULATEPOLAR_SRC_METHODS_VISCDRAGRAYMER_H_

#include <moduleBasics/runtimeIO.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/fuselage.h>

#include <memory>
#include <vector>
#include <string>

#include "../polar.h"

class viscDragRaymerInitialization {
 public:
    // geometry switches
    bool calculateWing;                 /**< calculate viscous drag of the wing switch */
    bool calculateFuselage;             /**< calculate viscous drag of the fuselage switch */
    bool calculateVTP;                  /**< calculate viscous drag of the VTP switch */
    bool calculateHTP;                  /**< calcualte viscous drag of the  HTP switch */
    bool calculateNacelles;             /**< calculate viscous drag of the nacelles switch */
    bool calculateMisc;                 /**< calculate miscellaneous drag switch */
    // transition parameters
    bool manualTransitionLocation;          /**< switch wether transition location is given manually or flow is considered turbulent over all surfaces */
    double transitionLocationWing;          /**< relative chord position on wing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationHTP;           /**< relative chord position on HTP, where the flow goes from laminar to turbulent [-] */
    double transitionLocationVTP;           /**< relative chord position on VTP, where the flow goes from laminar to turbulent [-] */
    // calibration parameters
    bool doCalibration;
    double CDvisc_CDsum;
    double CDvisc_CLfact;
    double CDvisc_CLexp;
    double CDvisc_CDsum_lowMa;
    double CDvisc_CLfact_lowMa;
    double CDvisc_CLexp_lowMa;
    // drag count correction
    bool doViscDragCtCorrForCalibration; /**< switch if viscous drag count correction should be applied [-] */
    double viscDragCtCorrForCalibrationWing; /**< correction of viscous wing drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationFuselage; /**< correction of viscous fuselage drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationHTP; /**< correction of viscous horizontal tailplane drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationVTP; /**< correction of viscous vertical tailplane drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationNacelles; /**< correction of viscous nacelle drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationMisc; /**< correction of viscous miscellaneous drag, given in drag counts [DC] */
    // other
    bool jetEngine;                         /**< switch if aircraft flies with jet engines or propellers */
    double the_delta_isa;                   /**< Temperature shift according to TLARS in [K] */
    // geometry
    geom2::MultisectionSurface<geom2::AirfoilSection> theWing;                      /**< geometry of the main wing */
    geom2::MultisectionSurface<geom2::AirfoilSection> theHTP;                       /**< geometry of the horzontal tail plane */
    geom2::MultisectionSurface<geom2::AirfoilSection> theVTP;                       /**< geometry of the vertical tail plaine */
    geom2::MultisectionSurface<geom2::PolygonSection> theFuselage;                  /**< geometry of the fuselage */
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> theNacelles;     /**< vector containing all nacelle geometries */
    viscDragRaymerInitialization();
    virtual ~viscDragRaymerInitialization() {}
};
class viscDragRaymer {
 public:
    // general variables
    std::string config;
    std::filesystem::path aircraftPath;
    std::string airfoilPath;
    // geometry switches
    bool calculateWing;                 /**< calculate viscous drag of the wing switch */
    bool calculateFuselage;             /**< calculate viscous drag of the fuselage switch */
    bool calculateVTP;                  /**< calculate viscous drag of the VTP switch */
    bool calculateHTP;                  /**< calcualte viscous drag of the  HTP switch */
    bool calculateNacelles;             /**< calculate viscous drag of the nacelles switch */
    bool calculateMisc;                 /**< calculate miscellaneous drag switch */
    // transition parameters
    bool manualTransitionLocation;
    double transitionLocationWing;          /**< relative chord position on wing, where the flow goes from laminar to turbulent [-] */
    double transitionLocationHTP;    /**< relative chord position on HTP, where the flow goes from laminar to turbulent [-] */
    double transitionLocationVTP;           /**< relative chord position on VTP, where the flow goes from laminar to turbulent [-] */
    // calibration parameters
    bool doCalibration;
    double CDvisc_CDsum;
    double CDvisc_CLfact;
    double CDvisc_CLexp;
    double CDvisc_CDsum_lowMa;
    double CDvisc_CLfact_lowMa;
    double CDvisc_CLexp_lowMa;
    // drag count correction
    bool doViscDragCtCorrForCalibration; /**< switch if viscous drag count correction should be applied [-] */
    double viscDragCtCorrForCalibrationWing; /**< correction of viscous wing drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationFuselage; /**< correction of viscous fuselage drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationHTP; /**< correction of viscous horizontal tailplane drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationVTP; /**< correction of viscous vertical tailplane drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationNacelles; /**< correction of viscous nacelle drag, given in drag counts [DC] */
    double viscDragCtCorrForCalibrationMisc; /**< correction of viscous miscellaneous drag, given in drag counts [DC] */
    // sonstiges
    bool jetEngine;                         /**< switch if aircraft flies with jet engines or propellers */
    double delta_isa;                       /**< temperature delta [K] */
    // geometry
    struct aircraftComponents {                                                         /**< struct which stores all possible aircraft geometry elements */
        geom2::MultisectionSurface<geom2::AirfoilSection> myWing;                      /**< geometry of the main wing */
        geom2::MultisectionSurface<geom2::AirfoilSection> myHTP;                       /**< geometry of the horzontal tail plane */
        geom2::MultisectionSurface<geom2::AirfoilSection> myVTP;                       /**< geometry of the vertical tail plaine */
        geom2::MultisectionSurface<geom2::PolygonSection> myFuselage;                  /**< geometry of the fuselage */
        std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> myNacelles;     /**< vector containing all nacelle geometries */
    };
    aircraftComponents theAircraftComponents;
    // geometry dependent wing parameters
    double Sref_Wing;           /**< Reference area of the wing [m^2] */
    double Swet_Wing;           /**< Wetted area of the wing [m^2] */
    double MAC_Wing;            /**< Mean aerodynamic chord of the wing [m] */
    double avgTtoC_Wing;        /**< Average thickness to chord ratio of the wing [-] */
    double mXtoC_Wing;          /**< Chordwise location of airfoil maximum thickness point [-] */
    double eta_ref_Wing;        /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_Wing;     /**< Local sweep at reference position [rad] */
    double QC_Wing;             /**< Form factor wing [-] */
    // geometry dependent HTP parameters
    double Swet_HTP;            /**< Wetted area of the HTP [m^2] */
    double MAC_HTP;             /**< Mean aerodynamic chord of the HTP [m] */
    double avgTtoC_HTP;         /**< Average thickness to chord ratio of the HTP [-] */
    double mXtoC_HTP;           /**< Chordwise location of airfoil maximum thickness point [-] */
    double eta_ref_HTP;         /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_HTP;      /**< Local sweep at reference position [rad] */
    double QC_HTP;              /**< Form factor HTP [-] */
    // geometry dependent VTP parameters
    double Swet_VTP;            /**< Wetted area of the VTP [m^2] */
    double MAC_VTP;             /**< Mean aerodynamic chord of the VTP [m] */
    double avgTtoC_VTP;         /**< Average thickness to chord ratio of the VTP [-] */
    double mXtoC_VTP;           /**< Chordwise location of airfoil maximum thickness point [-] */
    double eta_ref_VTP;         /**< Relative span position used for local sweep [-] */
    double phiAtmXtoC_VTP;      /**< Local sweep at reference position [rad] */
    double QC_VTP;              /**< Form factor VTP [-] */
    // geometry dependent fuselage parameters
    double Swet_Fuselage;       /**< Wetted area of the fuselage [m^2] */
    double l_Fuselage;          /**< lenght of the fuselage [m] */
    double maxWidthFuselage;    /**< maximum width of the fuselage [m] */
    double maxHeightFuselage;   /**< maximum height of the fuselage [m] */
    double QC_Fuselage;         /**< Form factor fuselage [-] */
    double fuselageUpsweepAngle;  /**< Up sweep angle of the fuselage tail [rad] */
    // geometry dependent nacelle parameters
    double Swet_Nacelles;       /**< Wetted area of one nacelle [m^2] */
    double l_Nacelles;          /**< Length of one nacelle [m] */
    double maxWidthNacelles;    /**< Maximum width of one nacelle [m] */
    double maxHeightNacelles;   /**< Maximum height of one nacelle [m] */
    double QC_Nacelles;         /**< Form factor nacelles [-] */
    double factorProtuberances; /**< protuberance factor of the aircraft [-] */
    // calculate parameters that are only dependent from the geometry
    void preprocessGeometry();
    // calculations
    /** \brief Overall function to calculate viscous drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationWing relative transition location where the flow turns turbulent on the wing
     *  \param transitionLocationHTP relative transition location where the flow turns turbulent on the HTP
     *  \param transitionLocationVTP relative transition location where the flow turns turbulent on the VTP
     */   
    std::vector <double> calculateViscDrag(const Polar& aPolar, const double transitionLocationWing,
                                    const double transitionLocationHTP, const double transitionLocationVTP);
    /** \brief Overall function to calculate viscous drag corrections
     *  \param componentDrags vector of the component drags which need to be corrected
     *  \param aMachNumber mach number for the calibration
     *  \param CL CL value for the calibration
     */                                       
    std::vector <double> calculateCorrectedViscDrag(const std::vector <double>& componentDrags, const double aMachNumber, const double CL);
    /** \brief Function to calculate the wing component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationWing relative transition location where the flow turns turbulent on the wing
     */          
    double calculateViscDragWing(const Polar& aPolar, const double transitionLocationWing);
    /** \brief Function to calculate the HTP component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationHTP relative transition location where the flow turns turbulent on the HTP
     */  
    double calculateViscDragHTP(const Polar& aPolar, const double transitionLocationHTP);
    /** \brief Function to calculate the VTP component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     *  \param transitionLocationVTP relative transition location where the flow turns turbulent on the VTP
     */  
    double calculateViscDragVTP(const Polar& aPolar, const double transitionLocationVTP);
    /** \brief Function to calculate the fuselage component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */  
    double calculateViscDragFuselage(const Polar& aPolar);
    /** \brief Function to calculate the nacelle component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */  
    double calculateViscDragNacelles(const Polar& aPolar);
    /** \brief Function to calculate the miscellaneous component drag
     *  \param aPolar polar/flight condition for which the drag gets calculated
     */     
    double calculateViscDragMisc(const Polar& aPolar);
    /** \brief Function to calculate the friction coefficient
     *  \param flightMach mach number
     *  \param ReynoldsNumber reynolds number
     */ 
    double getFrictionCoefficient(double flightMach, double ReynoldsNumber);
    explicit viscDragRaymer(const std::shared_ptr<RuntimeIO>& rtIO, const viscDragRaymerInitialization& viscDragInit);
    virtual ~viscDragRaymer() {}
};

#endif  // CALCULATEPOLAR_SRC_METHODS_VISCDRAGRAYMER_H_
