/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_METHODS_WAVEDRAGMASON_H_
#define CALCULATEPOLAR_SRC_METHODS_WAVEDRAGMASON_H_

#include "../polar.h"

#include <moduleBasics/runtimeIO.h>
#include <aircraftGeometry2/airfoil_surface.h>

#include <memory>
#include <vector>
#include <string>

class waveDragInitialization{
 public:
    double factorKa;                /**< airfoil technology factor [-]*/
    double maxAllowedSweep;         /**< maximum sweep angle where the method holds validity [deg] */
    size_t numberOfSegments;        /**< number of wing strips for the calculation */
    // calibration parameters
    double initialMachCruise;               /**< initial cruise mach number */
    bool useCostomReferenceAngleForSweep;   /**< switch if custom reference line position should be used */
    double customSweepPosition;             /**< relative position of chord where sweep angle for wave drag calculation should be calculated */
    bool doWaveDragMasonCalibration;        /**< switch if calibration function for wave drag should be applied */
    double CLFact;                          /**< factor value for the CL based calibration [-] */
    double CLExp;                           /**< exponent value for the CL based calibration [-] */
    bool doWaveDragMasonDragCountCorrection; /**< switch if drag count calibration for wave drag is applied */
    double waveDragCtCorrForCalibrationWing; /**< drag counts, that get added to or substracted from the wave drag of the wing */
    double waveDragCtCorrForCalibrationStabilizer; /**< drag counts, that get added to or substracted from the wave drag of the stabilizer */
    std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>> liftingSurfaces;
    geom2::MultisectionSurface<geom2::PolygonSection> theFuselage;
    waveDragInitialization();
    virtual ~waveDragInitialization() {}
};
class waveDragMason{
 public:
    double factorKa;                /**< airfoil technology factor [-]*/
    double maxAllowedSweep;         /**< maximum sweep angle where the method holds validity [deg] */
    size_t numberOfSegments;        /**< number of wing strips for the calculation */
    // calibration parameters
    double initialMachCruise;               /**< initial cruise mach number */
    bool useCostomReferenceAngleForSweep;   /**< switch if custom reference line position should be used */
    double customSweepPosition;             /**< relative position of chord where sweep angle for wave drag calculation should be calculated */
    bool doWaveDragMasonCalibration;        /**< switch if calibration function for wave drag should be applied */
    double CLFact;                          /**< factor value for the CL based calibration [-] */
    double CLExp;                           /**< exponent value for the CL based calibration [-] */
    bool doWaveDragMasonDragCountCorrection; /**< switch if drag count calibration for wave drag is applied */
    double waveDragCtCorrForCalibrationWing; /**< drag counts, that get added to or substracted from the wave drag of the wing */
    double waveDragCtCorrForCalibrationStabilizer; /**< drag counts, that get added to or substracted from the wave drag of the stabilizer */
    // geometry dependent parameters
    struct spanStationWing {    /**< struct for storing local thickness ratios of the wing */
        double thicknessRatio;  /**< thickness to chord ratio [-] */
        double span;            /**< span coordinate for the corresponding thickness ratio [m] */
    };
    /** \brief Function calculates the wave drag for a lifting surface at a polar point
     *  \param aPolarPoint the polar point for the calculation
     *  \param machNumber the mach number for the drag calculation
     *  \param liftingSurfaceIndex the index of the lifting surface, which gets calculated
     *  \param spanStations vector of span coordinates of the wing corresponding to the lift points
     *  \param Cl_distribution vector of lift coefficients corresponding to the span station coordinates
     */
    double calculateWaveDrag(const PolarPoint& aPolarPoint, const double machNumber, size_t liftingSurfaceIndex, const double chordWiseSweepPosition,
                                const std::vector<double>& spanStations, const std::vector<double>& Cl_distribution);
    /** \brief Function calculates the wave drag correction for a polar point using a clibration function
     *  \details this function should only be used for wing and not for stabilizer
     *  \param waveDragBeforeCorrection the value of the uncorrected wave drag point
     *  \param aPolarPoint the polar point for the calculation
     *  \param aMachNumber the mach number for the drag calculation
     */                                
    double calculateWaveDragCorrectionByCalibration(const double& waveDragBeforeCorrection, const PolarPoint& aPolarPoint, const double& aMachNumber);
    /** \brief Function calculates the wave drag correction for a polar point using the drag count calibration function
     *  \param waveDragCtCorrForCalibration the correction factor in drag counts [DC]
     */             
    double calculateWaveDragCorrectionByDragCounts(const double& waveDragCtCorrForCalibration);
    /** \brief Function which interpolates the thickness to chord ratio between lifting surface sections
     *  \param surfaceID the index of the surface which gets calculated
     *  \param yCoordinate the y coordinate for the interpolation
     */   
    double interpolateThicknessRatio(const size_t surfaceID, const double& yCoordinate);

    explicit waveDragMason(const std::shared_ptr<RuntimeIO>& rtIO, const waveDragInitialization& waveDragInit);
    virtual ~waveDragMason() {}

 private:
    const std::vector <geom2::MultisectionSurface<geom2::AirfoilSection>>& liftingSurfaces; /**< vector which contains all lifting surfaces */
    const geom2::MultisectionSurface<geom2::PolygonSection>& theFuselage;                   /**< const reference to aircraft fuselage geometry object */
    std::vector<std::vector<spanStationWing>> surfaceSectionThicknesses;                    /**< contains the thicknes ratios at all sections for the interpolation */
    bool tooLargeSweep;
};

#endif  // CALCULATEPOLAR_SRC_METHODS_WAVEDRAGMASON_H_
