# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def calculate_route_dependent_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate route dependent costs per year and depreciation period.

    Route dependent costs include all cost components that are directly attributable to flight operations. These
    include
        - fuel,
        - fees (handling, landing, air traffic control), and
        - maintenance.
    They are determined for one year and the entire depreciation period.

    This function adds the 'route_dependent_costs' dict to the 'dict_operating_costs' that contains the following
    parameters:
        - array route_dependent_costs_annual: Route dependent costs per year (range dependent) in EUR
        - array route_dependent_costs_depreciation_period: Route dependent costs per depreciation period (range
        dependent) in EUR
        - float route_dependent_costs_annual_design_point: Route dependent costs per year at design point for current
        aircraft in EUR
        - float route_dependent_costs_depreciation_period_design_point: Route dependent costs per depreciation period
        at design point for current aircraft in EUR

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    depreciation_period = dict_ac_data['depreciation_period']
    # Extract necessary variables from 'dict_operating_costs'.
    fuel_costs = dict_operating_costs['fuel_costs']['fuel_costs']
    handling_costs = dict_operating_costs['handling_costs']['handling_costs']
    idx_design_point = dict_operating_costs['payload_range_diagram_data']['idx_design_point']
    landing_costs = dict_operating_costs['landing_costs']['landing_costs']
    air_traffic_control_costs = dict_operating_costs['air_traffic_control_costs']['air_traffic_control_costs']
    maintenance_costs = dict_operating_costs['maintenance_costs']['maintenance_costs']

    # Sum up all cost shares to route dependent costs per year.
    route_dependent_costs_annual = (fuel_costs + handling_costs + landing_costs + air_traffic_control_costs
                                    + maintenance_costs)
    # Multiply route dependent costs with depreciation period.
    route_dependent_costs_depreciation_period = route_dependent_costs_annual * depreciation_period

    # Print.
    runtime_output.print('ROUTE DEPENDENT COSTS (per year): '
                         + f"{int(route_dependent_costs_annual[idx_design_point]):,}" + ' USD')
    runtime_output.print('                             *                            ')

    # Prepare output data.
    dict_operating_costs['route_dependent_costs'] = {}
    dict_operating_costs['route_dependent_costs']['route_dependent_costs_annual'] = route_dependent_costs_annual
    dict_operating_costs['route_dependent_costs']['route_dependent_costs_depreciation_period'] = \
        route_dependent_costs_depreciation_period
    dict_operating_costs['route_dependent_costs']['route_dependent_costs_annual_design_point'] = \
        route_dependent_costs_annual[idx_design_point]
    dict_operating_costs['route_dependent_costs']['route_dependent_costs_depreciation_period_design_point'] = \
        route_dependent_costs_depreciation_period[idx_design_point]

    return dict_operating_costs
