# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing functions for the preparation of user data."""


def user_method_data_input_preparation(routing_dict):
    """Prepare necessary input data for the user method from aircraft exchange and module configuration files.

    In this function, the user is responsible for preparing the data needed for the user method. Relevant general data
    are obtained from the aircraft exchange file and calculation specific parameter from the module configuration file.
    The user must submit the data in the following format:
        dict = {'parameter_name': [path to parameter node, expected data type], ...}

    :param dict routing_dict: Dictionary containing information on necessary data from module configuration file
    :returns:
        - dict data_to_extract_from_aircraft_exchange_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from aircraft exchange file
        - dict data_to_extract_from_module_configuration_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from module configuration file
    """

    """Aircraft exchange file."""
    # Paths.
    requirements_path = './requirements_and_specifications/requirements/top_level_aircraft_requirements/'
    transport_task_path = './requirements_and_specifications/design_specification/transport_task/'
    masses_and_cg_path = './analysis/masses_cg_inertia/'
    assessment_path = './assessment/'
    fuselage_path = './component_design/fuselage/'
    propulsion_path = './component_design/propulsion/'
    mission_path = './analysis/mission/'
    assessment_scenario_path = './requirements_and_specifications/assessment_scenario/'

    # Enter all parameters to be extracted from the aircraft exchange file.
    data_to_extract_from_aircraft_exchange_dict = {
        'altitude_cruise_design': [requirements_path + 'design_mission/initial_cruise_altitude', float],
        'altitude_cruise_study': [requirements_path + 'study_mission/initial_cruise_altitude', float],
        'm_cargo': [transport_task_path + 'cargo_definition/additional_cargo_mass', float],
        'm_luggage': [transport_task_path + 'passenger_definition/luggage_mass_per_passenger', float],
        'm_operating_empty': [masses_and_cg_path + 'operating_mass_empty/mass_properties/mass', float],
        'm_passenger': [transport_task_path + 'passenger_definition/mass_per_passenger', float],
        'm_payload_max': [masses_and_cg_path + 'maximum_payload_mass/mass_properties/mass', float],
        'm_payload_at_max_fuel': [assessment_path + 'performance/range/payload_maximum_fuel_at_maximum_takeoff_mass',
                                  float],
        'm_takeoff_max': [masses_and_cg_path + 'maximum_takeoff_mass/mass_properties/mass', float],
        'm_takeoff_design': [masses_and_cg_path + 'design_mass/mass_properties/mass', float],
        # 'm_takeoff_study': [masses_and_cg_path + 'study_mass/mass_properties/mass', float],
        'm_takeoff_study': [mission_path + 'study_mission/takeoff_mass', float],
        'mach_cruise_design': [requirements_path + 'design_mission/initial_cruise_mach_number', float],
        'mach_cruise_study': [requirements_path + 'study_mission/initial_cruise_mach_number', float],
        'n_cabin_crew_members': [fuselage_path + 'specific/geometry/fuselage[@ID="0"]/number_of_required_cabin_crew',
                                 float],
        'n_cockpit_crew_members': [fuselage_path
                                   + 'specific/geometry/fuselage[@ID="0"]/number_of_required_flight_crew', float],
        'passenger_class_distribution': [transport_task_path + 'passenger_class_definition/class_distribution', str],
        'n_passengers': [transport_task_path + 'passenger_definition/total_number_passengers', float],
        'range_at_max_fuel': [assessment_path + 'performance/range/range_max_fuel_at_maximum_takeoff_mass', float],
        'range_at_max_payload': [assessment_path + 'performance/range/range_max_payload_at_maximum_takeoff_mass',
                                 float],
        'range_ferry': [assessment_path + 'performance/range/range_maximum_fuel_empty', float],
        'static_thrust_per_engine': [propulsion_path + 'specific/propulsion[@ID="0"]/engine/SLST', float],
        'stage_length_design': [mission_path + 'design_mission/range', float],
        'stage_length_study': [mission_path + 'study_mission/range', float],
        'seat_load_factor_study':
            [requirements_path + 'study_mission/payload_fractions/passenger_mass_fraction', float],
        'flight_time_design': [mission_path + 'design_mission/flight_time', float],
        'flight_time_study': [mission_path + 'study_mission/flight_time', float],
        'flights_per_year': [assessment_scenario_path + 'flights_per_year', float]
    }

    """Module configuration file."""
    # Enter all general parameters to be extracted from the module configuration file. 'general parameters' means
    # parameters that do not differ according to the user layer. It should be noted that 'tmp_general' is only used to
    # shorten the path information in the 'general_data_to_extract_from_module_configuration_dict'.
    tmp_general = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                   + 'operating_cost_estimation_tu_berlin/general_direct_operating_costs_parameter')
    general_data_to_extract_from_module_configuration_dict = {
        'exchange_rate':
            [tmp_general + '/exchange_rate', float],
        'annual_lay_hours_maintenance':
            [tmp_general + '/flight_cycles/annual_lay_hours_maintenance', float],
        'annual_lay_hours_overhaul':
            [tmp_general + '/flight_cycles/annual_lay_hours_overhaul', float],
        'annual_lay_hours_reserve':
            [tmp_general + '/flight_cycles/annual_lay_hours_reserve', float],
        'air_traffic_control_price_factor_design':
            [tmp_general + '/air_traffic_control/air_traffic_control_price_factor_design', float],
        'air_traffic_control_price_factor_study':
            [tmp_general + '/air_traffic_control/air_traffic_control_price_factor_study', float],
        'airframe_repair_costs_per_flight':
            [tmp_general + '/maintenance/airframe_repair_costs_per_flight', float],
        'block_time_supplement_per_flight':
            [tmp_general + '/flight_cycles/block_time_supplement_per_flight', float],
        'cost_burden':
            [tmp_general + '/maintenance/cost_burden', float],
        'daily_night_curfew_time':
            [tmp_general + '/flight_cycles/daily_night_curfew_time', float],
        'depreciation_period':
            [tmp_general + '/capital/depreciation_period', float],
        'fees_handling':
            [tmp_general + '/handling/fees_handling', float],
        'fees_landing':
            [tmp_general + '/landing/fees_landing', float],
        'price_per_operating_empty_mass':
            [tmp_general + '/capital/price_per_operating_empty_mass', float],
        'rate_inflation':
            [tmp_general + '/miscellaneous/rate_inflation', float],
        'rate_insurance':
            [tmp_general + '/capital/rate_insurance', float],
        'rate_interest':
            [tmp_general + '/capital/rate_interest', float],
        'rate_labor':
            [tmp_general + '/maintenance/rate_labor', float],
        'revenue_per_freight_km_design':
            [tmp_general + '/related_direct_operating_costs/revenue_per_freight_km_design', float],
        'revenue_per_freight_km_study':
            [tmp_general + '/related_direct_operating_costs/revenue_per_freight_km_study', float]
    }

    # Enter all specific parameters to be extracted from the module configuration file. 'specific parameters' means
    # parameters that differ according to the user layer. It should be noted that 'tmp_specific' is only used to
    # shorten the path information in the 'specific_data_to_extract_from_module_configuration_dict'.
    tmp_specific = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                    + 'operating_cost_estimation_tu_berlin')
    specific_data_to_extract_from_module_configuration_dict = {
        'fuel_price': [tmp_specific + '/fuel_type[@ID="' + routing_dict['user_layer'] + '"]/fuel_price', float],
        'factor_engine_maintenance':
            [tmp_specific + '/fuel_type[@ID="' + routing_dict['user_layer'] + '"]/factor_engine_maintenance', float],
        'ratio_operating_empty_mass':
            [tmp_specific + '/fuel_type[@ID="' + routing_dict['user_layer'] + '"]/ratio_operating_empty_mass', float]
    }

    # Merge module configuration dictionaries.
    data_to_extract_from_module_configuration_dict = (general_data_to_extract_from_module_configuration_dict
                                                      | specific_data_to_extract_from_module_configuration_dict)

    return data_to_extract_from_aircraft_exchange_dict, data_to_extract_from_module_configuration_dict


def user_method_data_output_preparation(data_dict):
    """Prepare user-specific output data based on the calculation method results.

    This function is responsible for preparing the user-specific output data based on the results of the calculation
    method. The 'data_dict' input parameter contains the results of the module execution.
    The data for the key parameters output must be specified in the following format in order to be written correctly
    to the aircraft exchange file by the 'write_key_data_to_aircraft_exchange_file' function in the following step:
        dict = {'parameter_name': [path to parameter node, value, name (if needed)], ...}
    Important notes:
        (1) It should be noted that only key parameters may be written that have been previously defined by the module
        manager.
        (2) Attention must be paid to the proper path specifications, otherwise warnings may be issued or, in the worst
        case, errors may occur subsequently resulting in the write process and consequently the entire program being
        aborted.
        (3) If the path specifications contain IDs, these must start at '0' and be defined in ascending order without
        gaps.
    For the method-specific output, a path list and a dictionary is necessary to properly write the data to the
    method-specific XML file.
    the dictionary must be specified in the following format:
        dict = ...
    Note: If the user wants to export data from the design and study mission, two path lists and dictionaries are
    necessary.

    :param dict data_dict: Dictionary containing the results of the module execution
    :returns:
        - dict key_output_dict: Output dictionary containing key parameters that are written to aircraft XML file
        - dict method_specific_output_dict: Dictionary containing specific parameters that are written to
        method-specific output XML
    """

    """Key parameters output."""
    doc_path = './assessment/cost_estimation/operating_cost/direct_operating_cost/'

    key_output_dict = {
        'flights_per_year_design_mission':
            [doc_path + 'flights_per_year_design_mission',
             data_dict['design']['dependent_parameter']['flights_per_year']],
        # Direct operating costs shares.
        'route_independent_cost_annual':
            [doc_path + 'route_independent_cost_annual',
             data_dict['design']['route_independent_costs']['route_independent_costs_annual']],
        'route_dependent_cost_annual':
            [doc_path + 'route_dependent_cost_annual',
             data_dict['design']['route_dependent_costs']['route_dependent_costs_annual_design_point']],
        'direct_operating_cost_annual':
            [doc_path + 'direct_operating_cost_annual',
             data_dict['design']['direct_operating_costs']['direct_operating_costs_annual_design_point']]
    }
    
    # Check, if study mission exists.
    study_exists = ('study' in data_dict.keys())
    if study_exists:
        key_output_dict['flights_per_year_study_mission'] = ([doc_path + 'flights_per_year_study_mission', 2000])

    """Method-specific output (divided into design and study mission)."""
    # Define method-specific output dictionary.
    method_specific_output_dict = {}

    # Define specific output paths and dict for design mission.
    tmp_path = './calculation_results/operating_cost_estimation_tu_berlin/design_mission/'

    paths_to_specific_design_outputs_list = [
        tmp_path + 'direct_operating_costs/route_independent_cost_shares/capital_costs',
        tmp_path + 'direct_operating_costs/route_independent_cost_shares/crew_costs_per_year',
        tmp_path + 'direct_operating_costs/route_independent_cost_shares/crew_costs_per_depreciation_period',
        tmp_path + 'direct_operating_costs/route_independent_cost_shares/route_independent_costs_per_year',
        tmp_path
        + 'direct_operating_costs/route_independent_cost_shares/route_independent_costs_per_depreciation_period',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/fuel_costs',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/handling_costs',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/landing_costs',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/air_traffic_control_costs',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/maintenance_costs',
        tmp_path + 'direct_operating_costs/route_dependent_cost_shares/route_dependent_costs_per_year',
        tmp_path +
        'direct_operating_costs/route_dependent_cost_shares/route_dependent_costs_per_depreciation_period',
        tmp_path + 'direct_operating_costs/direct_operating_costs_per_year',
        tmp_path + 'direct_operating_costs/direct_operating_costs_per_depreciation_period',
        tmp_path + 'indirect_operating_costs'
    ]

    specific_output_dict_design = {
        'operating_cost_estimation_tu_berlin': {},
        'design_mission': {
            'attributes': {'description': 'Cost estimation results of the design mission'}},
        'direct_operating_costs': {
            'attributes': {'description': 'Direct operating costs'}},
        # Route independent cost shares.
        'route_independent_cost_shares': {
            'attributes': {'description': 'Route independent cost shares'}},
        'capital_costs': {
            'attributes': {'description': 'Capital costs per year'},
            'parameters': {'value': data_dict['design']['capital_costs']['capital_costs'],
                           'unit': 'USD'}},
        'crew_costs_per_year': {
            'attributes': {'description': 'Crew costs per year'},
            'parameters': {'value': data_dict['design']['crew_costs']['crew_costs_annual'],
                           'unit': 'USD'}},
        'crew_costs_per_depreciation_period': {
            'attributes': {'description': 'Crew costs per depreciation period at design point'},
            'parameters': {'value': data_dict['design']['crew_costs']['crew_costs_depreciation_period'],
                           'unit': 'USD'}},
        'route_independent_costs_per_year': {
            'attributes': {'description': 'Route independent costs per year at design point (sum of capital and crew '
                                          'costs)'},
            'parameters': {'value': data_dict['design']['route_independent_costs']['route_independent_costs_annual'],
                           'unit': 'USD'}},
        'route_independent_costs_per_depreciation_period': {
            'attributes': {'description': 'Route independent costs per depreciation period at design point (sum of '
                                          'capital and crew costs)'},
            'parameters': {'value': data_dict['design']['route_independent_costs'][
                'route_independent_costs_depreciation_period'],
                           'unit': 'USD'}},
        # Route dependent cost shares.
        'route_dependent_cost_shares': {
            'attributes': {'description': 'Route dependent cost shares'}},
        'fuel_costs': {
            'attributes': {'description': 'Fuel costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['fuel_costs']['fuel_costs_design_point'],
                'unit': 'USD'}},
        'handling_costs': {
            'attributes': {'description': 'Handling costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['handling_costs']['handling_costs'],
                'unit': 'USD'}},
        'landing_costs': {
            'attributes': {'description': 'Landing costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['landing_costs']['landing_costs'],
                'unit': 'USD'}},
        'air_traffic_control_costs': {
            'attributes': {'description': 'Air traffic control (ATC) costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['air_traffic_control_costs']['air_traffic_control_costs_design_point'],
                'unit': 'USD'}},
        'maintenance_costs': {
            'attributes': {'description': 'Maintenance costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['maintenance_costs']['maintenance_costs'],
                'unit': 'USD'}},
        'route_dependent_costs_per_year': {
            'attributes': {'description': 'Route dependent costs per year at design point'},
            'parameters': {
                'value': data_dict['design']['route_dependent_costs']['route_dependent_costs_annual_design_point'],
                'unit': 'USD'}},
        'route_dependent_costs_per_depreciation_period': {
            'attributes': {'description': 'Route dependent costs per depreciation period at design point'},
            'parameters': {
                'value': data_dict['design']['route_dependent_costs'][
                    'route_dependent_costs_depreciation_period_design_point'],
                'unit': 'USD'}},
        # Direct operating costs.
        'direct_operating_costs_per_year': {
            'attributes': {'description': 'Direct operating costs per year at design point (sum of route dependent and '
                                          'route independent costs)'},
            'parameters': {
                'value': data_dict['design']['direct_operating_costs']['direct_operating_costs_annual_design_point'],
                'unit': 'USD'}},
        'direct_operating_costs_per_depreciation_period': {
            'attributes': {'description': 'Direct operating costs per depreciation period at design point (sum of route'
                                          ' dependent and route independent costs)'},
            'parameters': {
                'value': data_dict['design']['direct_operating_costs'][
                    'direct_operating_costs_depreciation_period_design_point'],
                'unit': 'USD'}},
        # Indirect operating costs.
        'indirect_operating_costs': {
            'attributes': {'description': 'Indirect operating costs'},
            'parameters': {
                'value': data_dict['design']['indirect_operating_costs']['indirect_operating_costs'],
                'unit': 'USD'}},
        # List.
        'list_of_paths': paths_to_specific_design_outputs_list
    }

    method_specific_output_dict['specific_output_dict_design'] = specific_output_dict_design

    # Define specific output paths and dict for study mission.
    if study_exists:
        tmp_path = './calculation_results/operating_cost_estimation_tu_berlin/study_mission/'

        paths_to_specific_study_outputs_list = [
            tmp_path + 'direct_operating_costs/route_independent_cost_shares/capital_costs',
            tmp_path + 'direct_operating_costs/route_independent_cost_shares/crew_costs_per_year',
            tmp_path + 'direct_operating_costs/route_independent_cost_shares/crew_costs_per_depreciation_period',
            tmp_path + 'direct_operating_costs/route_independent_cost_shares/route_independent_costs_per_year',
            tmp_path
            + 'direct_operating_costs/route_independent_cost_shares/route_independent_costs_per_depreciation_period',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/fuel_costs',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/handling_costs',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/landing_costs',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/air_traffic_control_costs',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/maintenance_costs',
            tmp_path + 'direct_operating_costs/route_dependent_cost_shares/route_dependent_costs_per_year',
            tmp_path +
            'direct_operating_costs/route_dependent_cost_shares/route_dependent_costs_per_depreciation_period',
            tmp_path + 'direct_operating_costs/direct_operating_costs_per_year',
            tmp_path + 'direct_operating_costs/direct_operating_costs_per_depreciation_period',
            tmp_path + 'indirect_operating_costs'
        ]

        specific_output_dict_study = {
            'operating_cost_estimation_tu_berlin': {},
            'study_mission': {
                'attributes': {'description': 'Cost estimation results of the study mission'}},
            'direct_operating_costs': {
                'attributes': {'description': 'Direct operating costs'}},
            # Route independent cost shares.
            'route_independent_cost_shares': {
                'attributes': {'description': 'Route independent cost shares'}},
            'capital_costs': {
                'attributes': {'description': 'Capital costs per year'},
                'parameters': {'value': data_dict['study']['capital_costs']['capital_costs'],
                               'unit': 'USD'}},
            'crew_costs_per_year': {
                'attributes': {'description': 'Crew costs per year'},
                'parameters': {'value': data_dict['study']['crew_costs']['crew_costs_annual'],
                               'unit': 'USD'}},
            'crew_costs_per_depreciation_period': {
                'attributes': {'description': 'Crew costs per depreciation period at design point'},
                'parameters': {'value': data_dict['study']['crew_costs']['crew_costs_depreciation_period'],
                               'unit': 'USD'}},
            'route_independent_costs_per_year': {
                'attributes': {
                    'description': 'Route independent costs per year at design point (sum of capital and crew '
                                   'costs)'},
                'parameters': {
                    'value': data_dict['study']['route_independent_costs']['route_independent_costs_annual'],
                    'unit': 'USD'}},
            'route_independent_costs_per_depreciation_period': {
                'attributes': {'description': 'Route independent costs per depreciation period at design point (sum of '
                                              'capital and crew costs)'},
                'parameters': {'value': data_dict['study']['route_independent_costs'][
                    'route_independent_costs_depreciation_period'],
                               'unit': 'USD'}},
            # Route dependent cost shares.
            'route_dependent_cost_shares': {
                'attributes': {'description': 'Route dependent cost shares'}},
            'fuel_costs': {
                'attributes': {'description': 'Fuel costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['fuel_costs']['fuel_costs_design_point'],
                    'unit': 'USD'}},
            'handling_costs': {
                'attributes': {'description': 'Handling costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['handling_costs']['handling_costs'],
                    'unit': 'USD'}},
            'landing_costs': {
                'attributes': {'description': 'Landing costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['landing_costs']['landing_costs'],
                    'unit': 'USD'}},
            'air_traffic_control_costs': {
                'attributes': {'description': 'Air traffic control (ATC) costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['air_traffic_control_costs']['air_traffic_control_costs_design_point'],
                    'unit': 'USD'}},
            'maintenance_costs': {
                'attributes': {'description': 'Maintenance costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['maintenance_costs']['maintenance_costs'],
                    'unit': 'USD'}},
            'route_dependent_costs_per_year': {
                'attributes': {'description': 'Route dependent costs per year at design point'},
                'parameters': {
                    'value': data_dict['study']['route_dependent_costs']['route_dependent_costs_annual_design_point'],
                    'unit': 'USD'}},
            'route_dependent_costs_per_depreciation_period': {
                'attributes': {'description': 'Route dependent costs per depreciation period at design point'},
                'parameters': {
                    'value': data_dict['study']['route_dependent_costs'][
                        'route_dependent_costs_depreciation_period_design_point'],
                    'unit': 'USD'}},
            # Direct operating costs.
            'direct_operating_costs_per_year': {
                'attributes': {
                    'description': 'Direct operating costs per year at design point (sum of route dependent and '
                                   'route independent costs)'},
                'parameters': {
                    'value': data_dict['study']['direct_operating_costs'][
                        'direct_operating_costs_annual_design_point'],
                    'unit': 'USD'}},
            'direct_operating_costs_per_depreciation_period': {
                'attributes': {
                    'description': 'Direct operating costs per depreciation period at design point (sum of route'
                                   ' dependent and route independent costs)'},
                'parameters': {
                    'value': data_dict['study']['direct_operating_costs'][
                        'direct_operating_costs_depreciation_period_design_point'],
                    'unit': 'USD'}},
            # Indirect operating costs.
            'indirect_operating_costs': {
                'attributes': {'description': 'Indirect operating costs'},
                'parameters': {
                    'value': data_dict['study']['indirect_operating_costs']['indirect_operating_costs'],
                    'unit': 'USD'}},
            # List
            'list_of_paths': paths_to_specific_study_outputs_list
        }

        method_specific_output_dict['specific_output_dict_study'] = specific_output_dict_study

    return key_output_dict, method_specific_output_dict
