/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CREATE_MISSION_XML_SRC_LIBS_MISSION_STEPS_APPROACH_STEP_H_
#define CREATE_MISSION_XML_SRC_LIBS_MISSION_STEPS_APPROACH_STEP_H_

#include <string>
#include <aixml/endnode.h>

/** \class ApproachStep
 *  \brief Data needed fpr flight steps during approach
 */
class ApproachStep {
 public:
     /* Constructor and destructor */
     /** \brief Constructor for class ApproachStep
      * \param a_mode: reference to a flight mode of the approach step
      * \param a_rating: reference to the rating of the approach step
      * \param a_derate: derate of the approach step [-]
      * \param a_config: reference to configuration of the approach step
      * \param an_end_altitude: altitude at the end of the approach step [m]
      * \param a_rate_of_climb_limit: rate of climb of the approach step [m/s]
      * \param an_end_CAS: calibrated airspeed at the end of the approach step [m/s]
      * \param a_ground_distance: ground distance of the approach step [m]
      * \param a_glide_path: glide angle of the approach step [deg]
      */
    ApproachStep(const std::string &a_mode, const std::string &a_rating, double a_derate, const std::string &a_config, double an_end_altitude,
                 double a_rate_of_climb_limit, double an_end_CAS, double a_ground_distance, double a_glide_path);
    /** \brief Destructor for class ApproachStep
     */
    virtual ~ApproachStep();

    /** \brief function to write approach steps into mission file
     * \details It checks the given values and updates the endnodes within mission_file
     * \param subpath: path where the endnode shall write data
     * \param mission_file: xml node of the given mission file
     */
    void write_approach_steps(std::string subpath, node* mission_file);

 private:
    /* Attributes */
    Endnode<std::string> mode; /* flight mode of the approach step */
    Endnode<std::string> rating; /**< Rating of the departure step */
    Endnode<double> derate; /**< Derate of the departure step [-] */
    Endnode<std::string> configuration; /**< Configuration of the departure step */
    Endnode<double> rate_of_climb_limit; /**< Rate of climb of the departure step [m/s] */
    Endnode<double> end_CAS; /**< Calibrated air speed at the end of the departure step [m/s] */
    Endnode<double> end_altitude; /**< Flight altitude at the end of the departure step [m] */
    Endnode<double> ground_distance; /**< Ground distance of the approach step [m] */
    Endnode<double> glide_path; /**< Glide path of the approach step [deg] */
};

#endif // CREATE_MISSION_XML_SRC_LIBS_MISSION_STEPS_APPROACH_STEP_H_
