/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "mission.h"

void Mission::calculate_kerosene_emissions() {
    emissionsFlow.resize(number_of_steps);
    /* CO2 - Stoichiometric factor for the combustion of kerosene: 3.149 kg/kgFuel */
    this->LTO.CO2 = fuel_mass_LTO * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_CO2;;
    this->cruise.CO2 = fuel_mass_cruise * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_CO2;;
    this->LTO_Total.CO2 = this->LTO.CO2 * flights_per_lifecycle;
    this->cruiseTotal.CO2 = this->cruise.CO2 * flights_per_lifecycle;

    /* H2O - Stoichiometric factor for the combustion of kerosene: 1.2 kg/kgFuel */
    this->LTO.H2O = fuel_mass_LTO * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_H2O;;
    this->cruise.H2O = fuel_mass_cruise * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_H2O;;
    this->LTO_Total.H2O = this->LTO.H2O * flights_per_lifecycle;
    this->cruiseTotal.H2O = this->cruise.H2O * flights_per_lifecycle;

    /* SO2 - Stoichiometric factor for the combustion of kerosene: 0.84e-3 kg/kgFuel */
    this->LTO.SO2 = fuel_mass_LTO * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO2;
    this->cruise.SO2 = fuel_mass_cruise * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO2;
    this->LTO_Total.SO2 = this->LTO.SO2 * flights_per_lifecycle;
    this->cruiseTotal.SO2 = this->cruise.SO2 * flights_per_lifecycle;

    /* SO4 - Stoichiometric factor for the combustion of kerosene:2e-4 kg/kgFuel */
    this->LTO.SO4 = fuel_mass_LTO * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO4;
    this->cruise.SO4 = fuel_mass_cruise * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO4;
    this->LTO_Total.SO4 = this->LTO.SO4 * flights_per_lifecycle;
    this->cruiseTotal.SO4 = this->cruise.SO4 * flights_per_lifecycle;

    /* Fill mission vector with emissions proportional to fuel flow*/
    for (int16_t i(0); i < number_of_steps; i++) {
        emissionsFlow[i].CO2 = missionInfo[i].fuelFlow * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_CO2;
        emissionsFlow[i].H2O = missionInfo[i].fuelFlow * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_H2O;
        emissionsFlow[i].SO2 = missionInfo[i].fuelFlow * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO2;
        emissionsFlow[i].SO4 = missionInfo[i].fuelFlow * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO4;
    }

    /* HC - Non proportional to fuel flow */
    if (missionIOData->config.kerosene_method_HC == "mode_0") { //DLR Omega
        kerosene_HC_DLR();
    } else if (missionIOData->config.kerosene_method_HC == "mode_1") { //BFM2;
        kerosene_HC_BFM2();
    } else { //ICAO
        this->LTO.HC = (fuel_mass_taxi * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC))
                        + (fuel_mass_takeoff * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC))
                        + (fuel_mass_climbout * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC))
                        + (fuel_mass_approach * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC));
        this->cruise.HC = 0.;//not calculable with ICAO
    }
    this->LTO_Total.HC = this->LTO.HC * flights_per_lifecycle;
    this->cruiseTotal.HC = this->cruise.HC * flights_per_lifecycle;

    /* CO - Non proportional to fuel flow */
    if (missionIOData->config.kersosene_method_CO == "mode_0") { //DLR Omega
        kerosene_CO_DLR();
    } else if (missionIOData->config.kersosene_method_CO == "mode_1") { //BFM2;
        kerosene_CO_BFM2();
    } else { //ICAO
        this->LTO.CO = (fuel_mass_taxi * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO))
                        + (fuel_mass_takeoff * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO))
                        + (fuel_mass_climbout * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO))
                        + (fuel_mass_approach * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO));
        this->cruise.CO = 0.;//not calculable with ICAO
    }
    this->LTO_Total.CO = this->LTO.CO * flights_per_lifecycle;
    this->cruiseTotal.CO = this->cruise.CO * flights_per_lifecycle;

    /* NOx - Non proportional to fuel flow */
    if (missionIOData->config.kerosene_method_NOx == "mode_0") { //P3T3
        kerosene_NOx_P3T3();
    } else if (missionIOData->config.kerosene_method_NOx == "mode_1") { //BFM2;
        kerosene_NOx_BFM2();
    } else { //GasTurb
        kerosene_NOx_gasTurb();
    }
    this->LTO_Total.NOx = this->LTO.NOx * flights_per_lifecycle;
    this->cruiseTotal.NOx = this->cruise.NOx * flights_per_lifecycle;

    /* Soot - Non proportional to fuel flow */
    if (rtIOPt->aircraft_energy_carrier_type() == "kerosene") {
        if (missionIOData->config.kerosene_method_soot == "mode_0") { //DLR
            kerosene_soot_DLR();
        } else if (missionIOData->config.kerosene_method_soot == "mode_1") { //Whyte;
            kerosene_soot_whyte();
        } else { //const. EI = 0.025e-3 kg/kgFuel
            this->LTO.soot = fuel_mass_LTO * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_soot;
            this->cruise.soot = fuel_mass_cruise * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_soot;
        }
    }
    this->cruiseTotal.soot = this->cruise.soot * flights_per_lifecycle;
    this->LTO_Total.soot = this->LTO.soot * flights_per_lifecycle;

    /* Emissions calculated via ICAO Fuel and ICAO Times and ICAO EI (if available, otherwise constants) */
    double ICAO_LTO_Fuel =  fuel_mass_taxi_ICAO + fuel_mass_takeoff_ICAO + fuel_mass_climbout_ICAO + fuel_mass_approach_ICAO;
    this->ICAO_LTO.CO2 = ICAO_LTO_Fuel * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_CO2;
    this->ICAO_LTO.H2O = ICAO_LTO_Fuel * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_H2O;
    this->ICAO_LTO.SO2 = ICAO_LTO_Fuel * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_SO2;
    this->ICAO_LTO.HC = missionIOData->number_of_engines *
                        (fuel_mass_taxi_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC)
                        + fuel_mass_takeoff_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC));
    this->ICAO_LTO.CO = missionIOData->number_of_engines *
                        (fuel_mass_taxi_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO)
                        + fuel_mass_takeoff_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO));
    this->ICAO_LTO.NOx = missionIOData->number_of_engines *
                        (fuel_mass_taxi_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                        + fuel_mass_takeoff_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx)
                        + fuel_mass_approach_ICAO * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx));
    this->ICAO_LTO.soot = ICAO_LTO_Fuel * EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).emission_index_soot;
}

void Mission::kerosene_NOx_P3T3() {
    /* Taxiing NOx */
    this->LTO.NOx = missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx) * fuel_mass_taxi; //kg
    /* Non-taxiing NOx*/
    for (int mission_point = 0; mission_point < number_of_steps ; mission_point++) {
        double T3_mom = this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(mission_point).St3TotalPressure / pow(1. + (1.4 - 1.)
                    / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double n = 0.4; //exponent of the altitude correction for P3T3 Method, can be changed for best fit (0.4 recommended; p.29 Norman 2003)
        double beta = 7.90298 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(mission_point).Tamb)) + 3.00571 + 5.02808
                * log10(373.16 / (this->myFlightPath.pathPoints.at(mission_point).Tamb)) + 1.3816 * pow(10., -7.)
                * (1. - pow(10., 11.344 * (1. - (this->myFlightPath.pathPoints.at(mission_point).Tamb)
                / 373.16))) + 8.1328 * pow(10., -3.) * (pow(10., 3.49149 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(mission_point).Tamb))) - 1.);
        double P_v = 0.014504 * pow(10., beta); //psi
        //(psi = Pa * 0.00014504); AEM3 Correction Implemented!!!
        double omega = 0.62198 * missionIOData->config.relative_humidity * P_v / (this->myFlightPath.pathPoints.at(mission_point).Pamb *
            0.00014504 - 0.37802 * missionIOData->config.relative_humidity * P_v);
        double H = -19.0 * (omega - 0.00634);
        double EINOx_SL = quadratic_fit(T3_mom, T3_ID, T3_TO, T3_CL, T3_AP,
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx), 0);// [kg/kg]
        double P3_SL = potential_fit(T3_mom, T3_ID, T3_TO, T3_CL, T3_AP, P3_ID, P3_TO, P3_CL, P3_AP, 1); //[kPa]
        if (EINOx_SL < 0.) {
            EINOx_SL = 0.;
        }
        EINOxKerosene.push_back(EINOx_SL * pow(P3_mom / P3_SL, n) * exp(H)); //kg/kg
        emissionsFlow[mission_point].NOx = EINOxKerosene.back() * missionInfo[mission_point].fuelFlow;

        /* calculate NOx mass */
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(mission_point) * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.NOx += EINOxKerosene.at(mission_point) * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
    }
}

void Mission::kerosene_NOx_BFM2() {
    this->LTO.NOx = missionIOData->number_of_engines * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                    * missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxi_time; //kg
    for (int mission_point = 0; mission_point < number_of_steps ; mission_point++) {
        double delta_amb = this->myFlightPath.pathPoints.at(mission_point).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(mission_point).Tamb / 288.15;
        double beta = 7.90298 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(mission_point).Tamb)) + 3.00571 + 5.02808 * log10(373.16
            / (this->myFlightPath.pathPoints.at(mission_point).Tamb)) + 1.3816 * pow(10., -7.) * (1. - pow(10., 11.344
            * (1. - (this->myFlightPath.pathPoints.at(mission_point).Tamb) / 373.16))) + 8.1328 * pow(10., -3.)
            * (pow(10., 3.49149 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(mission_point).Tamb))) - 1.);
        double P_v = 0.014504 * pow(10., beta);
        //(psi = Pa * 0.00014504); AEM3 Correction Implemented!!!
        double omega = 0.62198 * missionIOData->config.relative_humidity * P_v / (this->myFlightPath.pathPoints.at(mission_point).Pamb *
            0.00014504 - 0.37802 * missionIOData->config.relative_humidity * P_v);
        double H = -19.0 * (omega - 0.0063);
        ///For EINOx: Choose between linear_least_square_fit() or point_to_point_fit()
        EINOxKerosene.push_back(linear_least_square_fit(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx),
                missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx),
                missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx),
                missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx), mission_point)
                * exp(H) * pow(pow(delta_amb, 1.02) / pow(theta_amb, 3.3), 0.5)); //[kg/kg];
        emissionsFlow[mission_point].NOx = EINOxKerosene.back() * missionInfo[mission_point].fuelFlow;
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(mission_point) * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.NOx += EINOxKerosene.at(mission_point) * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
    }
}

void Mission::kerosene_NOx_gasTurb() {
    this->LTO.NOx = missionIOData->number_of_engines * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                    * missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxi_time; //kg
    for (int mission_point = 1; mission_point < number_of_steps; mission_point++) {
        /* First mission point */
        double altitude(missionInfo.at(mission_point-1).altitude); // Flight Level [m]
        double Mach(missionInfo.at(mission_point-1).Mach); // Mach number
        double N1 =  missionInfo.at(mission_point-1).N1; // Current low pressure spool speed [-]
        missionIOData->myEngine.set_operating_point({N1, altitude, Mach});
        double NOx_EI_1(missionIOData->myEngine.get_engine_NOx_emission_index());
        /* Second mission point */
        altitude = missionInfo.at(mission_point).altitude;// Flight Level [m]
        Mach = missionInfo.at(mission_point).Mach;// Mach number
        N1 =  missionInfo.at(mission_point-1).N1; // Current low pressure spool speed [-]
        missionIOData->myEngine.set_operating_point({N1, altitude, Mach});
        double NOx_EI_2(missionIOData->myEngine.get_engine_NOx_emission_index());
        /* Average EI */
        EINOxKerosene.push_back((NOx_EI_2 + NOx_EI_1) / 2.);
        /* Nox masses */
        time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(mission_point-1) * this->myFlightPath.pathPoints.at(mission_point-1).fuelFlow * time_step; //kg
        } else {
            this->cruise.NOx += EINOxKerosene.at(mission_point-1) * this->myFlightPath.pathPoints.at(mission_point-1).fuelFlow * time_step; //kg
        }
        /* Nox flow */
        emissionsFlow[mission_point].NOx = EINOxKerosene.back() * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_HC_BFM2() {
    this->LTO.HC = missionIOData->number_of_engines * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC)
                * missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxi_time; //kg
    for (int mission_point = 0; mission_point < number_of_steps ; mission_point++) {
        double delta_amb = this->myFlightPath.pathPoints.at(mission_point).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(mission_point).Tamb / 288.15;
        double EIHC = bi_linear_least_square_fit(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC), mission_point)
                        * pow(theta_amb, 3.3) / pow(delta_amb, 1.02); //[kg/kg];
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.HC += EIHC * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.HC += EIHC * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].HC = EIHC * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_HC_DLR() {
    double c = 0.4; //exponent of the altitude correction for Omega Method, can be changed for best fit (0.4 recommended p.39 Petzold 2002)
    //Pressure in bar according to Doepelheuer 2002! but no matter, because potential fit!
    double omegaVC_ID = St3MassFlow_ID / pow(P3_ID / 100., 1.8) / exp(T3_ID / 300.);
    double omegaVC_TO = St3MassFlow_TO / pow(P3_TO / 100., 1.8) / exp(T3_TO / 300.);
    double omegaVC_CL = St3MassFlow_CL / pow(P3_CL / 100., 1.8) / exp(T3_CL / 300.);
    double omegaVC_AP = St3MassFlow_AP / pow(P3_AP / 100., 1.8) / exp(T3_AP / 300.);
    this->LTO.HC = missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC) * fuel_mass_taxi; //kg
    for (int mission_point = 1; mission_point < number_of_steps ; mission_point++) {
        double T3_mom = this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(mission_point).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double omegaVC_mom = this->myFlightPath.pathPoints.at(mission_point).St3MassFlow / pow(P3_mom / 100., 1.8) / exp(T3_mom / 300.);
        missionIOData->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(mission_point).N1, 0, 0});
        double EIHC_SL = quadratic_fit(omegaVC_mom, omegaVC_ID, omegaVC_TO, omegaVC_CL, omegaVC_AP,
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC), 1);
        double T3_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
        double P3t_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));
        double EIHC = EIHC_SL * pow(this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature / T3_SL * P3_SL / P3_mom, c); //kg/kg
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.HC += EIHC_SL * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else if (this->myFlightPath.pathPoints.at(mission_point).altitude < 12192.) {
            this->cruise.HC += EIHC_SL * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.HC += EIHC * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].HC = EIHC * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_CO_BFM2() {
    this->LTO.CO = missionIOData->number_of_engines * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO)
                    * missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxi_time; //kg
    for (int mission_point = 0; mission_point < number_of_steps ; mission_point++) {
        double delta_amb = this->myFlightPath.pathPoints.at(mission_point).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(mission_point).Tamb / 288.15;
        double EICO = bi_linear_least_square_fit(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO),
                        missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO), mission_point)
                        * pow(theta_amb, 3.3) / pow(delta_amb, 1.02); //[kg/kg];
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.CO += EICO * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.CO += EICO * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].CO = EICO * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_CO_DLR() {
    double c = 0.4; //exponent of the altitude correction for Omega Method, can be changed for best fit (0.4 recommended p.39 Petzold 2002)
    //Pressure in bar according to Doepelheuer 2002! but no matter, because potential fit!
    double omegaVC_ID = St3MassFlow_ID / pow(P3_ID / 100., 1.8) / exp(T3_ID / 300.);
    double omegaVC_TO = St3MassFlow_TO / pow(P3_TO / 100., 1.8) / exp(T3_TO / 300.);
    double omegaVC_CL = St3MassFlow_CL / pow(P3_CL / 100., 1.8) / exp(T3_CL / 300.);
    double omegaVC_AP = St3MassFlow_AP / pow(P3_AP / 100., 1.8) / exp(T3_AP / 300.);
    this->LTO.CO = missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO) * fuel_mass_taxi; //kg
    for (int mission_point = 1; mission_point < number_of_steps ; mission_point++) {
        double T3_mom = this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(mission_point).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double omegaVC_mom = this->myFlightPath.pathPoints.at(mission_point).St3MassFlow / pow(P3_mom / 100., 1.8) / exp(T3_mom / 300.);
        double EICO_SL = quadratic_fit(omegaVC_mom, omegaVC_ID, omegaVC_TO, omegaVC_CL, omegaVC_AP,
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO),
                            missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO), 1); // [kg/kg]
        //Set temporary engine rating to SL
        missionIOData->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(mission_point).N1, 0, 0});
        //Use manual engine rating for sea-level engine data
        double T3_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
        double P3t_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));
        double EICO = EICO_SL * pow(this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature / T3_SL * P3_SL / P3_mom, c); //kg/kg
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.CO += EICO_SL * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else if (this->myFlightPath.pathPoints.at(mission_point).altitude < 12192.) {
            this->cruise.CO += EICO_SL * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        } else {
            this->cruise.CO += EICO * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].CO = EICO * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_soot_DLR() {
    double EI_soot = 0.; //[kg/kg]
    //DLR Correlation
    double c_sootRef_ID = 0.0000026156 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_TO = 0.0000026156 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 3.)
                             + 0.0022367 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 3.)
                             + 0.10955 * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_CL = 0.0000026156 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_AP = 0.0000026156 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN) + 0.0012842; //mg/kg
    EI_soot = c_sootRef_ID * St3MassFlow_ID / (fuelFlow_ID / missionIOData->number_of_engines) / 1.225 / 1000000.; //kg/kg
    this->LTO.soot = EI_soot * fuel_mass_taxi; //kg
    for (int mission_point = 1; mission_point < number_of_steps ; mission_point++) {
        double fuelFlow_mom = this->myFlightPath.pathPoints.at(mission_point).fuelFlow;//[kg/s]
        double T3_mom = this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature;//[K]
        double P3_mom = this->myFlightPath.pathPoints.at(mission_point).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double ro_air_mom = this->myFlightPath.pathPoints.at(mission_point).Pamb / this->myFlightPath.pathPoints.at(mission_point).Tamb / 286.7;//[kg/m^3]
        double AFR = this->myFlightPath.pathPoints.at(mission_point).St3MassFlow / fuelFlow_mom;
        double T_flame_mom = 2281.*pow(P3_mom / 100., 0.009375) + 0.000178 * pow(P3_mom / 100, 0.055)
                             * (this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature - 298.);//[K]
        double c_sootRef_mom = linear_interpolation((fuelFlow_mom / missionIOData->number_of_engines),
                                            (fuelFlow_ID / missionIOData->number_of_engines), (fuelFlow_TO / missionIOData->number_of_engines),
                                            (fuelFlow_CL / missionIOData->number_of_engines),
                                            (fuelFlow_AP / missionIOData->number_of_engines), c_sootRef_ID, c_sootRef_TO, c_sootRef_CL, c_sootRef_AP);//[mg/m^3]
        //Set temporary engine rating to SL
        missionIOData->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(mission_point).N1, 0, 0});
        //Use manual engine rating for sea-level engine data
        double P3t_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));//[kPa]
        double T_flameSL = 2281.*pow(P3_mom / 100., 0.009375) + 0.000178 * pow(P3_mom / 100., 0.055) * (T3_mom - 298.);//[K]
        double c_soot_alt = c_sootRef_mom * pow((P3_mom / P3_SL), 1.35) * exp(-20000 / T_flame_mom) / exp(-20000 / T_flameSL); //mg/kg
        EI_soot = c_soot_alt * AFR / ro_air_mom / 1000000.;//kg/kg
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.soot += EI_soot * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
            if (c_soot_alt > c_soot_LTO_max)
                c_soot_LTO_max = c_soot_alt;
        } else {
            this->cruise.soot += EI_soot * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].soot = EI_soot * missionInfo[mission_point].fuelFlow;
    }
}

void Mission::kerosene_soot_whyte() {
    /**< \todo [T2150] (https://unicado.ilr.rwth-aachen.de/T2150): Correct Soot Correlation
     */
    double EI_soot = 0.;//[kg/kg]
    EI_soot = 1.08 * pow(10., 27.) * pow(P3_ID / 1000. / (St3MassFlow_ID / (fuelFlow_ID / missionIOData->number_of_engines)
              * 167. / 28.85 * 0.42 / 12.), 2.7) / pow(1.91667, 5.45) / pow(T3_ID, 8.66) / 1000.;
    this->LTO.soot = EI_soot * fuel_mass_taxi; //kg
    for (int mission_point = 1; mission_point < number_of_steps ; mission_point++) {
        double AFR = this->myFlightPath.pathPoints.at(mission_point).St3MassFlow / (this->myFlightPath.pathPoints.at(mission_point).fuelFlow / missionIOData->number_of_engines);
        double OCRatio = AFR * 167. / 28.85 * 0.42 / 12.;
        if (OCRatio < 11.)
            OCRatio = 11.;
        //myRuntimeInfo->debug<< "AFR = " << AFR << std::endl;
        //myRuntimeInfo->debug<< "X = " << X << std::endl;
        double P3_mom = this->myFlightPath.pathPoints.at(mission_point).St3TotalPressure / pow(1. + (1.4 - 1.)
                / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.),
                 1.4 / (1.4 - 1.)) / pow(1. + (1.4 - 1.) / 2.*pow(this->myFlightPath.pathPoints.at(mission_point).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double ro_air_mom = this->myFlightPath.pathPoints.at(mission_point).Pamb / this->myFlightPath.pathPoints.at(mission_point).Tamb / 286.7;//[kg/m^3]
        //EI_soot = 1.08*pow(10.,29.)*pow(P3_mom/1000./X,2.7) / pow(1.91667,5.45) / pow(this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature,8.66)/1000.;
        EI_soot = 1.08 * pow(10., 27.) * pow(P3_mom / 1000. / OCRatio, 2.7) / pow(1.91667, 5.45)
                 / pow(this->myFlightPath.pathPoints.at(mission_point).St3StaticTemperature, 8.66) / 1000.; //kg/kg
        double c_soot_alt = EI_soot * ro_air_mom / AFR * 1000000.;//[mg/m^3]
        //myRuntimeInfo->debug<< "EI_soot = " << EI_soot << std::endl;
        if (mission_point == 0) {
            time_step = missionInfo.at(mission_point).time;
        } else {
            time_step = missionInfo.at(mission_point).time - missionInfo.at(mission_point - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(mission_point).altitude <= 914.4) {
            this->LTO.soot += EI_soot * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
            if (c_soot_alt > c_soot_LTO_max)
                c_soot_LTO_max = c_soot_alt;
        } else {
            this->cruise.soot += EI_soot * this->myFlightPath.pathPoints.at(mission_point).fuelFlow * time_step; //kg
        }
        emissionsFlow[mission_point].soot = EI_soot * missionInfo[mission_point].fuelFlow;
    }
}
