/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "flopsMassMethod.h"

#include <unitConversion/unitConversion.h>

#include <cmath>

namespace flops {

double fin_mass(double technology_factor, int number_of_fins, double area, double taper_ratio, double design_gross_weight) {
  const double dg_lb = convertUnit(KILO, GRAM, NOPREFIX, POUND, design_gross_weight);
  const double area_sqft = convertUnit(AREA, NOPREFIX, METER, NOPREFIX, FOOT, area);
  const double mass = 0.32 * pow(dg_lb, 0.3) * (taper_ratio + 0.5) * area_sqft;
  return technology_factor * convertUnit(NOPREFIX, POUND, KILO, GRAM, mass);
}

double horizontal_stabilizer_mass(double technology_factor, double area, double taper_ratio, double design_gross_weight) {
  const double dg_lb = convertUnit(KILO, GRAM, NOPREFIX, POUND, design_gross_weight);
  const double area_sqft = convertUnit(AREA, NOPREFIX, METER, NOPREFIX, FOOT, area);
  const double mass = 0.53 * pow(dg_lb, 0.2) * (taper_ratio + 0.5) * area_sqft;
  return technology_factor * convertUnit(NOPREFIX, POUND, KILO, GRAM, mass);
}

double vertical_stabilizer_mass(double technology_factor, int number_of_tails, double area, double taper_ratio, double design_gross_weight) {
  const double dg_lb = convertUnit(KILO, GRAM, NOPREFIX, POUND, design_gross_weight);
  const double area_sqft = convertUnit(AREA, NOPREFIX, METER, NOPREFIX, FOOT, area);
  const double mass = 0.32 * pow(dg_lb, 0.3) * (taper_ratio + 0.5) * pow(number_of_tails, 0.7) * pow(area_sqft, 0.85);
  return technology_factor * convertUnit(NOPREFIX, POUND, KILO, GRAM, mass);
}
}  // namespace flops
