# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import os
import sys
import json
import importlib

# Import own modules.
from src.blended_wing_body.empirical.blended_wing_body_design_tu_berlin.general.call_functions.estimatemasses import estimate_masses  # noPep8 e501
from src.blended_wing_body.empirical.blended_wing_body_design_tu_berlin.general.plot_functions.plotpayloadtubes import plot_payload_tubes  # noPep8 e501
from src.blended_wing_body.empirical.blended_wing_body_design_tu_berlin.general.call_functions.prepareexternaldata import prepare_external_data  # noPep8 e501
from src.blended_wing_body.empirical.blended_wing_body_design_tu_berlin.general.call_functions.estimatepositionsandcog import estimate_positions_and_cog  # noPep8 e501


def blended_wing_body_design(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """  Module providing calculation functions for passenger tube and cabin design for blended wing body configurations
         in accordance to CS 25.801ff.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict payload_tube_description: Dict containing parameter and values of current fuselage design
    """
    ''' Initialize local parameter '''
    convergence_flag = False
    payload_tube_description = {'passenger_cabin_width': {},
                                'interior_description': {},
                                'payload_tube_wall_thickness': {},
                                'payload_tube_width': {},
                                'total_width_of_passenger_compartments': {},
                                'payload_tube_floor_thickness': {},
                                'required_cabin_length_per_tube': {},
                                'payload_tube_center_reference_points': {},
                                'payload_tube_wall_reference_points': {},
                                'payload_tube_length': {},
                                'passenger_deck_volume': {},
                                'passenger_deck_height': {},
                                'cargo_deck_volume': {},
                                'cargo_deck_height': {},
                                'geometry_flag': False
                                }
    word_to_insert_after = 'generate'
    files_to_remove = ['prepareexternaldata.py', 'estimatemasses.py', 'estimatepositionsandcog.py']

    ''' Read and prepare external data for fuselage and cabin design '''
    # Call function to prepare external data.
    dict_ac_exchange = prepare_external_data(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config,
                                             runtime_output)

    ''' Run fuselage design '''
    # List all entries in the directory
    all_entries = os.listdir(paths_and_names['working_directory'] + '/' +
                             routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')

    # Filter out only files (ignore directories)
    call_function_list = [entry for entry in all_entries
                          if os.path.isfile(paths_and_names['working_directory'] + '/' +
                                            routing_dict['module_import_name'].replace('.', '/')
                                            + '/general/call_functions' + '/' + entry)]

    # Remove unnecessary files from call function list.
    for file in files_to_remove:
        if file in call_function_list:
            call_function_list.remove(file)

    call_function_list = sorted(call_function_list, key=lambda x: (int(x.split('_')[1]), x))

    i = 1
    while not convergence_flag and i < 100:
        runtime_output.print('Current iteration of fuselage design is: ' + str(i))
        # Loop across all python files from call_function directory to execute them one by one.
        for python_module in call_function_list:
            try:
                # Import own modules from call function directory.
                import_call_function = (routing_dict['module_import_name']
                                        + '.general.call_functions.' + os.path.splitext(python_module)[0])

                if 'generate' in import_call_function:
                    position_to_insert = python_module.find(word_to_insert_after) + len(word_to_insert_after)
                    python_module = word_to_insert_after + '_' + python_module[position_to_insert:]

                    function_name = os.path.splitext(python_module)[0]
                    # Import calculation module.
                    module = importlib.import_module(import_call_function)

                    # Call function depending on routing parameter.
                    payload_tube_description = getattr(module, function_name)(dict_ac_exchange, dict_mod_config,
                                                                              payload_tube_description, runtime_output)

            # Exception handling for module import error.
            except ModuleNotFoundError as module_import_error:
                runtime_output.critical('Error: ' + str(module_import_error) + ' found in '
                                        + routing_dict['module_name'] + '. Program aborted.')
                sys.exit('Exit information: The own python module to be imported could not be found!')

        ''' convergence check '''
        # check if the fuselage geometry flag is true
        #  -> if true: -> check convergence -> otherwise restart landing gear estimation
        if payload_tube_description['geometry_flag']:
            convergence_flag = True
            runtime_output.print('Payload tube design for blended wing body configuration '
                                 'successful after ' + str(i) + ' iterations!')  # noPep8 e501

        i += 1

    # Call function to estimate fuselage and cabin masses.
    #payload_tube_description = estimate_masses(dict_ac_exchange, dict_mod_config, payload_tube_description, runtime_output)

    # Call function to estimate positions and center of gravity.
    #payload_tube_description = estimate_positions_and_cog(dict_ac_exchange, dict_mod_config, payload_tube_description, runtime_output)

    ''' generate plot of payload tubes for visualization '''
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))
    if plot_switch:
        plot_payload_tubes(paths_and_names, payload_tube_description, runtime_output)

    ''' debug print '''
    runtime_output.debug('Debug: The blended_wing_body_design function was successfully executed!')

    return payload_tube_description
