# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing functions for data postprocessing."""
# Import standard modules.

# Import own modules.
from datapostprocessingmodule import method_data_postprocessing
from datapostprocessingmodule import write_key_data_to_aircraft_exchange_file
from datapostprocessingmodule import prepare_element_tree_for_module_key_parameter


def data_postprocessing(paths_and_names, routing_dict, data_dict, runtime_output):
    """Perform data postprocessing and write data to an aircraft exchange file.

    This function is responsible for data postprocessing that involves the following steps:
        (1) Data preparation: The module manager prepares a list containing all paths to the key parameters that must
        be written to the aircraft exchange file.
        (2) Write data to the aircraft exchange file: The results of the module execution are passed on to the function
        responsible for properly writing the data to the aircraft exchange file.
        (3) Method-specific data postprocessing: User-defined method-specific postprocessing is conducted as needed.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing the result of the module execution (direct operating cost estimation)
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """

    """Data preparation."""
    # Changes to this list are the sole responsibility of the module manager!
    paths_to_key_parameters_list = [
        # total fuselage parameter paths
        './component_design/fuselage/position/x',
        './component_design/fuselage/position/y',
        './component_design/fuselage/position/z',
        './component_design/fuselage/mass_properties/mass',
        './component_design/fuselage/mass_properties/inertia/j_xx',
        './component_design/fuselage/mass_properties/inertia/j_yy',
        './component_design/fuselage/mass_properties/inertia/j_zz',
        './component_design/fuselage/mass_properties/inertia/j_xy',
        './component_design/fuselage/mass_properties/inertia/j_xz',
        './component_design/fuselage/mass_properties/inertia/j_yx',
        './component_design/fuselage/mass_properties/inertia/j_yz',
        './component_design/fuselage/mass_properties/inertia/j_zx',
        './component_design/fuselage/mass_properties/inertia/j_zy',
        './component_design/fuselage/mass_properties/center_of_gravity/x',
        './component_design/fuselage/mass_properties/center_of_gravity/y',
        './component_design/fuselage/mass_properties/center_of_gravity/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/name',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/position/x',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/position/y',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/position/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/direction/x',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/direction/y',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/direction/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/mass',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_xx',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_yy',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_zz',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_xy',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_xz',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_yx',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_yz',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_zx',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/inertia/j_zy',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/center_of_gravity/x',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/center_of_gravity/y',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_properties/center_of_gravity/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/name',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/section_shape',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/origin/x',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/origin/y',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/origin/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/upper_height',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/lower_height',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/sections/section[@ID="0"]/width',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/number_of_required_cabin_crew',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/number_of_required_flight_crew',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_structure/component_mass[@ID="0"]/name',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_structure/component_mass[@ID="0"]/mass',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_furnishing/component_mass[@ID="0"]/name',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_furnishing/component_mass[@ID="0"]/mass',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_operator_items/component_mass[@ID="0"]/name',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/mass_breakdown/fuselage_operator_items/component_mass[@ID="0"]/mass',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/position/x',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/position/y',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/position/z',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/mass_properties/mass',
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/mass_properties/center_of_gravity/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/mass_properties/center_of_gravity/y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/mass_properties/center_of_gravity/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/name',  # noPep8 e501# noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/front_reference_points/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/front_reference_points/y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/front_reference_points/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/front_reference_points/upper_z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/front_reference_points/lower_z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/aft_reference_points/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/aft_reference_points/y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/aft_reference_points/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/aft_reference_points/upper_z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_reference_points/aft_reference_points/lower_z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/front_reference_points/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/front_reference_points/left_y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/front_reference_points/right_y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/front_reference_points/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/aft_reference_points/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/aft_reference_points/left_y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/aft_reference_points/right_y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_wall_reference_points/aft_reference_points/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_structural_wall_thickness',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_tube_water_volume',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/name',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/position/x',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/position/y',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/position/z',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/payload_deck_structural_floor_thickness',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/payload_deck_water_volume',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/payload_deck_length',  # noPep8 e501
        './component_design/fuselage/specific/geometry/fuselage[@ID="0"]/fuselage_accommodation/payload_tube[@ID="0"]/payload_decks/payload_deck[@ID="0"]/payload_deck_required_galley_power'  # noPep8 e501
        ]

    module_key_parameters_dict = {
        # total fuselage parameter
        'component_design': {
            'fuselage': {
                'attributes': {
                    'description': 'Geometric description of the aircraft fuselage.',
                    'tool_level': '0'},
                'position': {
                    'attributes': {'description':
                                   'Position of the fuselages with regard to the global reference point.'},
                    'x': {
                        'attributes': {'description': 'Distance in x direction with regard to the global reference '
                                                      'point. (fuselage nose point)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-100',
                        'upper_boundary': '100'},
                    'y': {
                        'attributes': {'description': 'Distance in y direction with regard to the global reference '
                                                      'point. (fuselage nose point)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-25',
                        'upper_boundary': '25'},
                    'z': {
                        'attributes': {'description': 'Distance in z direction with regard to the global reference '
                                                      'point. (fuselage nose point)'},
                        'value': '0',
                        'unit': 'm',
                        'lower_boundary': '-5',
                        'upper_boundary': '5'},
                },
                'mass_properties': {
                    'attributes': {
                        'description': 'Mass properties of all fuselages.'},
                    'mass': {
                        'attributes': {'description': 'Mass of all fuselages.'},
                        'value': '0',
                        'unit': 'kg',
                        'lower_boundary': '0',
                        'upper_boundary': 'inf'},
                    'inertia': {
                        'attributes': {
                            'description': 'Inertia of all fuselages with regard to the total center of gravity.'},
                        'j_xx': {
                            'attributes': {
                                'description': 'Inertia of the of all fuselages in x.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yy': {
                            'attributes': {
                                'description': 'Inertia of the of all fuselages in y.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zz': {
                            'attributes': {
                                'description': 'Inertia of the of all fuselages in z.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_xy': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in xy.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_xz': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in xz.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yx': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in yx.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_yz': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in yz.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zx': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in zx.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                        'j_zy': {
                            'attributes': {
                                'description': 'Inertia of all fuselages in zy.'},
                            'value': '0',
                            'unit': 'kgm^2',
                            'lower_boundary': '-inf',
                            'upper_boundary': 'inf'},
                    },
                    'center_of_gravity': {
                        'attributes': {
                            'description': 'Center of gravity of all fuselages.'},
                        'x': {
                            'attributes': {
                                'description': 'Center of gravity in x-direction with regard to the global reference '
                                               'point. (of all fuselages)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '0',
                            'upper_boundary': '50'},
                        'y': {
                            'attributes': {
                                'description': 'Center of gravity in y-direction with regard to the global reference '
                                               'point. (of all fuselages)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '-5',
                            'upper_boundary': '5'},
                        'z': {
                            'attributes': {
                                'description': 'Center of gravity in z-direction with regard to the global reference '
                                               'point. (of all fuselages)'},
                            'value': '0',
                            'unit': 'm',
                            'lower_boundary': '-5',
                            'upper_boundary': '5'}
                    }
                },
                'specific': {
                    'attributes': {
                        'description': 'Fuselage design specific output parameter.'},
                    'geometry': {
                        'attributes': {
                            'description': 'Geometry description of the entire fuselage design.'},
                        'fuselage': {
                            'attributes': {
                                'ID': '0',
                                'description': 'Geometrical description of one entire fuselage.'},
                            'name': {
                                'attributes': {
                                    'description': 'Name of the fuselage.'},
                                'value': 'center_fuselage'},
                            'position': {
                                'attributes': {
                                    'description': 'Position of one entire fuselage with regard to the fuselage global '
                                                   'position.'},
                                'x': {
                                    'attributes': {
                                        'description': 'Distance in x direction with regard to the fuselage gear global'
                                                       ' position.'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-100',
                                    'upper_boundary': '100'},
                                'y': {
                                    'attributes': {
                                        'description': 'Distance in y direction with regard to the fuselage gear global'
                                                       ' position.'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-25',
                                    'upper_boundary': '25'},
                                'z': {
                                    'attributes': {
                                        'description': 'Distance in z direction with regard to the fuselage gear global'
                                                       ' position.'},
                                    'value': '0',
                                    'unit': 'm',
                                    'lower_boundary': '-5',
                                    'upper_boundary': '5'}
                            },
                            'direction': {
                                'attributes': {
                                    'description': 'Unit vector according to global coordinate system for direction '
                                                   'applied at position.'},
                                'x': {
                                    'attributes': {'description': 'X component of the unit vector.'},
                                    'value': '1',
                                    'unit': '1',
                                    'lower_boundary': '0',
                                    'upper_boundary': '1'},
                                'y': {
                                    'attributes': {'description': 'Y component of the unit vector.'},
                                    'value': '0',
                                    'unit': '1',
                                    'lower_boundary': '0',
                                    'upper_boundary': '1'},
                                'z': {
                                    'attributes': {'description': 'Z component of the unit vector.'},
                                    'value': '0',
                                    'unit': '1',
                                    'lower_boundary': '0',
                                    'upper_boundary': '1'},
                            },
                            'mass_properties': {
                                'attributes': {
                                    'description': 'Mass properties of one entire fuselage.'},
                                'mass': {
                                    'attributes': {'description': 'Mass of one entire fuselage.'},
                                    'value': '0',
                                    'unit': 'kg',
                                    'lower_boundary': '0',
                                    'upper_boundary': '100000'},
                                'inertia': {
                                    'attributes': {
                                        'description': 'Inertia of one entire fuselage with regard to his '
                                                       'center of gravity.'
                                    },
                                    'j_xx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in x.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in y.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in z.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_xy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in xy.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_xz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in xz.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in yx.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_yz': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in yz.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zx': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in zx.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                    'j_zy': {
                                        'attributes': {
                                            'description': 'Inertia of one entire fuselage in zy.'},
                                        'value': '0',
                                        'unit': 'kgm^2',
                                        'lower_boundary': '-inf',
                                        'upper_boundary': 'inf'},
                                },
                                'center_of_gravity': {
                                    'attributes': {
                                        'description': 'Center of gravity of one entire fuselage.'},
                                    'x': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global fuselage center of gravity position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '50'},
                                    'y': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global fuselage center of gravity position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-25',
                                        'upper_boundary': '25'},
                                    'z': {
                                        'attributes': {
                                            'description': 'Center of gravity in x-direction with regard to the '
                                                           'global fuselage center of gravity position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-5',
                                        'upper_boundary': '5'}
                                }
                            },
                            'sections': {
                                'attributes': {
                                    'description': 'Geometrical description of the fuselage sections of one entire '
                                                   'fuselage.'},
                                'section': {
                                    'attributes': {
                                        'ID': '0',
                                        'description': 'Geometrical description of one fuselage section.'},
                                    'name': {
                                        'attributes': {
                                            'description': 'Name of the fuselage section.'},
                                        'value': 'section_1'},
                                    'section_shape': {
                                        'attributes': {
                                            'description':
                                                'Contains a string with the path to the shape *.dat file '
                                                'or the key word: ellipse.'},
                                        'value': 'ellipse'},
                                    'origin': {
                                        'attributes': {
                                            'description': 'Origin of fuselage section (local).'},
                                        'x': {
                                            'attributes': {
                                                'description': 'Distance in x direction with regard to the global '
                                                               'reference point.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-75',
                                            'upper_boundary': '75'},
                                        'y': {
                                            'attributes': {
                                                'description': 'Distance in y direction with regard to the global '
                                                               'reference point.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-75',
                                            'upper_boundary': '75'},
                                        'z': {
                                            'attributes': {
                                                'description': 'Distance in z direction with regard to the global '
                                                               'reference point.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-75',
                                            'upper_boundary': '75'}
                                    },
                                    'upper_height': {
                                        'attributes': {
                                            'description': 'Height of the upper half of the fuselage section.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '10'},
                                    'lower_height': {
                                        'attributes': {
                                            'description': 'Height of the lower half of the fuselage section.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '10'},
                                    'width': {
                                        'attributes': {
                                            'description': 'Width of the fuselage section.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '10'}
                                }
                            },
                            'number_of_required_cabin_crew': {
                                'attributes': {
                                    'description': 'Number of minimum required flight attendants.'},
                                'value': '0',
                                'unit': '1',
                                'lower_boundary': '0',
                                'upper_boundary': '100'
                            },
                            'number_of_required_flight_crew': {
                                'attributes': {
                                    'description': 'Number of minimum required flight crew members.'},
                                'value': '0',
                                'unit': '1',
                                'lower_boundary': '1',
                                'upper_boundary': '4'
                            },
                            'mass_breakdown': {
                                'attributes': {
                                    'description': 'Mass breakdown of fuselage design module in accordance to CPACS.'},
                                'fuselage_structure': {
                                    'attributes': {
                                        'description': 'Mass breakdown of the fuselage structure block.'},
                                    'component_mass': {
                                        'attributes': {
                                            'ID': '0',
                                            'description': 'Mass of a single component of the fuselage structure '
                                                           'block.'},
                                        'name': {
                                            'attributes': {
                                                'description': 'Name of the fuselage structure component.'},
                                            'value': 'fuselage_structure_mass'},
                                        'mass': {
                                            'attributes': {
                                                      'description': 'Mass of the fuselage structure component.'},
                                            'value': '0',
                                            'unit': 'kg',
                                            'lower_boundary': '0',
                                            'upper_boundary': '100000'}
                                    }
                                },
                                'fuselage_furnishing': {
                                    'attributes': {
                                        'description': 'Mass breakdown of the fuselage furnishing block.'},
                                    'component_mass': {
                                        'attributes': {
                                            'ID': '0',
                                            'description': 'Mass of a single component of the fuselage furnishing '
                                                           'block.'},
                                        'name': {
                                            'attributes': {
                                                'description': 'Name of the fuselage furnishing component.'},
                                            'value': 'cabin_lightnings'},
                                        'mass': {
                                            'attributes': {
                                                'description': 'Mass of the fuselage furnishing component.'},
                                            'value': '0',
                                            'unit': 'kg',
                                            'lower_boundary': '0',
                                            'upper_boundary': '100000'}
                                    }
                                },
                                'fuselage_operator_items': {
                                    'attributes': {
                                        'description': 'Mass breakdown of the fuselage  operator items block.'},
                                    'component_mass': {
                                        'attributes': {
                                            'ID': '0',
                                            'description': 'Mass of a single component of the fuselage  operator items '
                                                           'block.'},
                                        'name': {
                                            'attributes': {
                                                'description': 'Name of the fuselage operator items component.'},
                                            'value': 'catering_mass'},
                                        'mass': {
                                            'attributes': {
                                                'description': 'Mass of the fuselage operator item component.'},
                                            'value': '0',
                                            'unit': 'kg',
                                            'lower_boundary': '0',
                                            'upper_boundary': '100000'}
                                    }
                                }
                            },
                            'fuselage_accommodation': {
                                'attributes': {
                                    'description': 'Description of the accommodation of current fuselage.'},
                                'position': {
                                    'attributes': {
                                        'description': 'Position of the payload tubes with regard to the fuselage '
                                                       'position.'},
                                    'x': {
                                        'attributes': {
                                            'description': 'Distance in x direction with regard to the fuselage '
                                                           'position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-100',
                                        'upper_boundary': '100'},
                                    'y': {
                                        'attributes': {
                                            'description': 'Distance in y direction with regard to the fuselage '
                                                           'position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-0',
                                        'upper_boundary': '0'},
                                    'z': {
                                        'attributes': {
                                            'description': 'Distance in z direction with regard to the fuselage '
                                                           'position.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '-5',
                                        'upper_boundary': '5'},
                                },
                                'mass_properties': {
                                    'attributes': {
                                        'description': 'Mass properties of the payload tubes of one entire fuselage.'},
                                    'mass': {
                                        'attributes': {'description': 'Mass of the payload tubes of one entire '
                                                                      'fuselage.'},
                                        'value': '0',
                                        'unit': 'kg',
                                        'lower_boundary': '0',
                                        'upper_boundary': '100000'},
                                    'center_of_gravity': {
                                        'attributes': {
                                            'description': 'Center of gravity of all payload tubes.'},
                                        'x': {
                                            'attributes': {
                                                'description': 'Center of gravity in x-direction with regard to the '
                                                               'center of gravity of the fuselage.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '0',
                                            'upper_boundary': '50'},
                                        'y': {
                                            'attributes': {
                                                'description': 'Center of gravity in y-direction with regard to the '
                                                               'center of gravity of the fuselage.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-5',
                                            'upper_boundary': '5'},
                                        'z': {
                                            'attributes': {
                                                'description': 'Center of gravity in z-direction with regard to the '
                                                               'center of gravity of the fuselage.'},
                                            'value': '0',
                                            'unit': 'm',
                                            'lower_boundary': '-5',
                                            'upper_boundary': '5'}
                                    }
                                },
                                'payload_tube': {
                                    'attributes': {
                                        'ID': '0',
                                        'description': 'Geometrical description of one payload tube of the fuselage.'},
                                    'name': {
                                        'attributes': {
                                            'description': 'Name of the payload tube.'},
                                        'value': 'center_payload_tube'},
                                    'payload_tube_reference_points': {
                                        'attributes': {
                                            'description': 'Payload tube center reference points in x, y and '
                                                           'z-direction referred to fuselage nose point.'},
                                        'front_reference_points': {
                                            'attributes': {
                                                'description': 'Reference points in the front of payload tube.'},
                                            'x': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in x-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-100',
                                                'upper_boundary': '100'},
                                            'y': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '35'},
                                            'z': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '15'},
                                            'upper_z': {
                                                'attributes': {
                                                    'description': 'Upper payload tube reference point in '
                                                                   'z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '15'},
                                            'lower_z': {
                                                'attributes': {
                                                    'description': 'Lower payload tube reference point in '
                                                                   'z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '0'}
                                        },
                                        'aft_reference_points': {
                                            'attributes': {
                                                'description': 'Reference points in the aft of payload tube.'},
                                            'x': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in x-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-100',
                                                'upper_boundary': '100'},
                                            'y': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '35'},
                                            'z': {
                                                'attributes': {
                                                    'description': 'Payload tube reference point in z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '15'},
                                            'upper_z': {
                                                'attributes': {
                                                    'description': 'Upper payload tube reference point in '
                                                                   'z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '15'},
                                            'lower_z': {
                                                'attributes': {
                                                    'description': 'Lower payload tube reference point in '
                                                                   'z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '0'}
                                        }
                                    },
                                    'payload_tube_wall_reference_points': {
                                        'attributes': {
                                            'description': 'Payload tube wall reference points in x, y and z-direction'
                                                           ' referred to fuselage nose point.'},
                                        'front_reference_points': {
                                            'attributes': {
                                                'description': 'Reference points in the front of payload tube.'},
                                            'x': {
                                                'attributes': {
                                                    'description': 'Wall reference point in x-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-100',
                                                'upper_boundary': '100'},
                                            'left_y': {
                                                'attributes': {
                                                    'description': 'Left wall reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-35',
                                                'upper_boundary': '35'},
                                            'right_y': {
                                                'attributes': {
                                                    'description': 'Right wall reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-35',
                                                'upper_boundary': '35'},
                                            'z': {
                                                'attributes': {
                                                    'description': 'Wall reference point in z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '15'}
                                        },
                                        'aft_reference_points': {
                                            'attributes': {
                                                'description': 'Reference points in the aft of payload tube.'},
                                            'x': {
                                                'attributes': {
                                                    'description': 'Wall reference point in x-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-100',
                                                'upper_boundary': '100'},
                                            'left_y': {
                                                'attributes': {
                                                    'description': 'Left wall reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-35',
                                                'upper_boundary': '35'},
                                            'right_y': {
                                                'attributes': {
                                                    'description': 'Right wall reference point in y-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-35',
                                                'upper_boundary': '35'},
                                            'z': {
                                                'attributes': {
                                                    'description': 'Wall reference point in z-direction.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '-15',
                                                'upper_boundary': '15'}
                                        }
                                    },
                                    'payload_tube_structural_wall_thickness': {
                                        'attributes': {'description': 'Structural wall thickness of the payload tube.'},
                                        'value': '0',
                                        'unit': 'm',
                                        'lower_boundary': '0',
                                        'upper_boundary': '1'
                                    },
                                    'payload_tube_water_volume': {
                                        'attributes': {'description': 'Total water volume of one entire payload tube.'},
                                        'value': '0',
                                        'unit': 'm^3',
                                        'lower_boundary': '0',
                                        'upper_boundary': '1000'
                                    },
                                    'payload_decks': {
                                        'attributes': {'description': 'Geometrical description of the payload decks.'},
                                        'payload_deck': {
                                            'attributes': {
                                                'ID': '0',
                                                'description': 'Geometrical description of the payload decks in one'
                                                               ' payload tube.'},
                                            'name': {
                                                'attributes': {
                                                    'description': 'Name of the payload deck.'},
                                                'value': 'passenger_deck'},
                                            'position': {
                                                'attributes': {
                                                    'description': 'Position of the payload deck with regard to the'
                                                                   ' payload tube reference point.'},
                                                'x': {
                                                    'attributes': {
                                                        'description': 'Distance in x direction with regard to the'
                                                                       ' payload tube reference point.'},
                                                    'value': '0',
                                                    'unit': 'm',
                                                    'lower_boundary': '-100',
                                                    'upper_boundary': '100'},
                                                'y': {
                                                    'attributes': {
                                                        'description': 'Distance in y direction with regard to the '
                                                                       'payload tube reference point.'},
                                                    'value': '0',
                                                    'unit': 'm',
                                                    'lower_boundary': '-0',
                                                    'upper_boundary': '0'},
                                                'z': {
                                                    'attributes': {
                                                        'description': 'Distance in z direction with regard to the '
                                                                       'payload tube reference point.'},
                                                    'value': '0',
                                                    'unit': 'm',
                                                    'lower_boundary': '-5',
                                                    'upper_boundary': '5'},
                                            },
                                            'payload_deck_structural_floor_thickness': {
                                                'attributes': {'description': 'Structural floor thickness of current '
                                                                              'payload deck.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '1'
                                            },
                                            'payload_deck_water_volume': {
                                                'attributes': {'description': 'Total water volume of the payload'
                                                                              ' deck.'},
                                                'value': '0',
                                                'unit': 'm^3',
                                                'lower_boundary': '0',
                                                'upper_boundary': '1000'
                                            },
                                            'payload_deck_length': {
                                                'attributes': {'description': 'Total length of the payload deck.'},
                                                'value': '0',
                                                'unit': 'm',
                                                'lower_boundary': '0',
                                                'upper_boundary': '100'
                                            },
                                            'payload_deck_required_galley_power': {
                                                'attributes': {'description': 'Required galley power of the payload '
                                                                              'deck.'},
                                                'value': '0',
                                                'unit': 'W',
                                                'lower_boundary': '0',
                                                'upper_boundary': '100000'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    paths_and_names = prepare_element_tree_for_module_key_parameter(paths_and_names, module_key_parameters_dict)

    # Run 'user_method_data_output_preparation' from 'usermethoddatapreparation.py'.
    key_output_dict, method_specific_output_dict = routing_dict['func_user_method_data_output_preparation'](data_dict)
    # Extract tool level from routing dictionary.
    tool_level = routing_dict['tool_level']

    """Write data to aircraft exchange file."""
    # Extract root and path to aircraft exchange file.
    root_of_aircraft_exchange_tree = paths_and_names['root_of_aircraft_exchange_tree']
    path_to_aircraft_exchange_file = paths_and_names['path_to_aircraft_exchange_file']
    # Write key data to aircraft exchange file.
    write_key_data_to_aircraft_exchange_file(root_of_aircraft_exchange_tree, path_to_aircraft_exchange_file,
                                             paths_to_key_parameters_list, key_output_dict, tool_level, runtime_output)

    """Method-specific postprocessing."""
    # Run 'method_data_postprocessing' from 'datapostprocessingmodule'.
    data_dict['module_key_parameters_dict'] = module_key_parameters_dict
    data_dict['paths_to_key_parameters_list'] = paths_to_key_parameters_list
    method_data_postprocessing(paths_and_names, routing_dict, data_dict,
                               method_specific_output_dict, runtime_output)
