# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing calculation functions provided by the user."""
# Import standard modules.
import os
import sys

# Import own modules.
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.fuselagedesign import fuselage_design  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.useexistinggeometry import use_existing_geometry  # noPep8 e501


def method_kerosene(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Fuselage design and cabin arrangement for kerosene-powered aircraft.

    This function performs fuselage design for kerosene-powered aircraft configurations.
    The fuselage and cabin arrangement is designed in accordance with CS.25.
    The output dictionary 'kerosene_output_dict' contains the results of the fuselage design module and
    is structured according to the following scheme:
        kerosene_output_dict = {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                 ...}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict kerosene_output_dict: Dictionary containing results from calculation for kerosene-powered aircraft
    """

    # initialize local parameter
    kerosene_output_dict = {}

    # check which sizing mode is selected -> if no valid mode is selected -> raise an error and abort program
    if dict_mod_config['program_mode'] == 'mode_0' or dict_mod_config['program_mode'] == 'mode_1':
        # check if sizing mode "mode_0" is selected -> if true: -> perform fuselage design estimation
        if dict_mod_config['program_mode'] == 'mode_0':
            kerosene_output_dict = fuselage_design(paths_and_names, routing_dict, dict_ac_exchange,
                                                   dict_mod_config, runtime_output)

        # Check if sizing mode "mode_1" is selected
        #  -> if true: -> perform retrofit fuselage design from existing geometry
        if dict_mod_config['program_mode'] == 'mode_1':
            runtime_output.print('Start reading of existing geometry from file.')
            # If the script is compiled, sys._MEIPASS gives the temp directory for PyInstaller.
            if getattr(sys, 'frozen', False):    
                # ToDo Pfad anpassen, wenn als exe ausgeführt
                path_to_existing_geometry_file = \
                    sys._MEIPASS + '/' + dict_mod_config['path_to_existing_geometry_file']
                print(paths_and_names['project_directory'] + os.path.realpath(dict_mod_config['path_to_existing_geometry_file']))

            # Check if the system prefix is different from the base prefix -> code is runnig in a vitual environment
            elif hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
                # ToDo Pfad anpassen, wenn als exe ausgeführt
                path_to_existing_geometry_file = \
                    os.path.join(paths_and_names['working_directory'], 
                                 os.path.relpath(dict_mod_config['path_to_existing_geometry_file']).replace(os.sep, '/'))
                print(path_to_existing_geometry_file)
                
            # If running as a regular Python script, use the current directory.
            else:
                path_to_existing_geometry_file = \
                    os.path.join(paths_and_names['project_directory'], 
                                 os.path.relpath(dict_mod_config['path_to_existing_geometry_file'])).replace(os.sep, '/')
                
            # Check if the file containing the geometrical description of the fuselage is existing
            #  -> if true: -> perform fuselage design with existing geometry
            if os.path.isfile(path_to_existing_geometry_file):
                dict_mod_config['path_to_existing_geometry_file'] = path_to_existing_geometry_file
                # Call function to perform fuselage design from existing geometry.
                kerosene_output_dict = use_existing_geometry(paths_and_names, routing_dict, dict_ac_exchange,
                                                             dict_mod_config, runtime_output)
            # Else condition: File containing the geometrical description of fuselage does not exist
            #  -> abort program.
            else:
                runtime_output.critical('Error: Program mode "use existing geometry" is selected but the file: "'
                                        + dict_mod_config['path_to_existing_geometry_file'] + '" could not be found. \n'
                                        '                                            Program aborted!')
                sys.exit('Exit information: Missing existing fuselage geometry file!')

    # else condition: no valid sizing mode is selected -> program aborted.
    else:
        runtime_output.critical('Error: No valid program mode for fuselage design selected! '
                                '\nPlease check the module configuration file. Program Aborted!')
        sys.exit('Exit information: No valid program mode is selected!')

    return kerosene_output_dict
