/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef INITIALSIZING_SRC_TAWINITIALSIZINGDATA_H_  // NOLINT
#define INITIALSIZING_SRC_TAWINITIALSIZINGDATA_H_

#include <tuple>
#include <map>
#include <memory>
#include <string>
#include <vector>

namespace fs = std::filesystem;

/**
 * \brief storage class for acxml data and updating
*/
class TawInitialSizingData {
 public:
/* Cabin and Cargo */
    /** \brief contains all cabin properties
     */
    class cabin {
     public:

        int numberPassengers = 0;/**< number of passengers*/
        double massPassenger = 0.;/**< mass per passengers*/
        double massLuggage = 0.;/**< Luggage mass per passengers*/
        double massCargo = 0.;/**< additional cargo mass*/
    };
    cabin Accommodation; /**contains all cabin properties read from the acxml*/

/* Class Distribution */
    /** \brief contains all mission properties
     */
    class classParameters {
     public:
       float President = 0.;
       float First = 0.;
       float Business = 0.;
       float PremiumEconomy = 0.;
       float Economy = 0.;
    };
    classParameters ClassDistribution;

/* Mission Parameters */
    /** \brief contains all mission properties
     */
    class missionParameters {
     public:
        double machNumber = 0.; /**< initial cruise mach number*/
        double altitude = 0.;  /**< initial cruise altitude*/
        double vApproach = 0.; /**< approach velocity*/
        double sTOFL = 0.;
        double range = 0.;/**< mission range*/
        double distAlter = 0.;
        double loitertime = 0.;
        double distalt = 0.;
        double contingency = 0.;/**< percentage of the reserved fuel*/
        double Gamma = 0.;
    
    };
    missionParameters MissionInfo;

/* Energy Carrier Parameters */
    /** \brief contains all mission properties
     */
    class carrierParameters {
     public:
        std::string energyCarrier; /**< energy carrier, i.e, kerosene, liquid hydrogen*/
        
    };
    carrierParameters EnergyCarrier;
    
    /* FUNCTIONS FOR WRITING TO THE ACXML FILE */

    /** 
     * @struct NodeInfo
     * @brief Struct which contains the name, id, and description of a new node
     * when inserting the node into the tree.
     */

    struct NodeInfo {
        std::string name{}; /**< The name of the node. */
        std::string id{}; /**< The id of the surface. */
        std::string description{}; /**< The description of the node. */
    };

    /**
     * @brief Create a endnode object and set its properties.
     * 
     * @tparam T The value type of the endnode.
     * @param value The value of the endnode.
     * @param unit The unit of the endnode.
     * @param node_path The location in the parent node of the endnode.
     * @param description The description of the endnode.
     * @return Endnode<T> Return the created endnode.
     */
    template<typename T>
    auto create_endnode(const T value, const std::string& node_path, const std::string& description) -> Endnode<T> {
        /* Create the node an set its properties */
        Endnode<T> new_node{node_path, description};
        new_node.set_value(value);
        /* Return the new node */
        return new_node;
    };


    /**
     * @brief Create a endnode object and set its properties.
     * 
     * @tparam T The value type of the endnode.
     * @param value The value of the endnode.
     * @param unit The unit of the endnode.
     * @param node_path The location in the parent node of the endnode.
     * @param description The description of the endnode.
     * @return Endnode<T> Return the created endnode.
     */
    template<typename T>
    auto create_endnode(const T value, const std::string& unit, const std::string& node_path, const std::string& description) -> Endnode<T> {
        /* Create the node an set its properties */
        Endnode<T> new_node{node_path, description};
        new_node.set_value(value);
        new_node.set_unit(unit);
        /* Return the new node */
        return new_node;
    };


};

#endif // INITIALSIZING_SRC_TAWINITIALSIZINGDATA_H_  // NOLINT
