# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import os
import sys

# Import own modules.
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.flexible_pavement.initializeacnflexparameter import initialize_acn_flex_parameter  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.flexible_pavement.acnflexcomp import acn_flex_comp  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.flexible_pavement.setacnflexinputparameter import set_acn_flex_input_parameter  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.plot_functions.generateacnplots import generate_acn_plots  # noPep8 e501


def estimate_aircraft_classification_number(paths_and_names, landing_gear_description, dict_ac_exchange,
                                            dict_mod_config, runtime_output):
    """ This function estimates the aircraft classification number in accordance to ICAO COMFAA tool.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_acn_flex_parameter: Dict containing parameter and values of current acn estimation
    """
    # initialize parameter to estimate ACN for flex pavement
    dict_acn_flex_parameter, dict_constants = initialize_acn_flex_parameter(runtime_output)

    # set acn flex specific parameter to dictionary
    dict_acn_flex_parameter = set_acn_flex_input_parameter(paths_and_names, landing_gear_description, dict_ac_exchange,
                                                           dict_mod_config, runtime_output, dict_acn_flex_parameter,
                                                           dict_constants)

    # start calculation of ACN for flex pavement
    dict_acn_flex_parameter = acn_flex_comp(dict_acn_flex_parameter, dict_constants, runtime_output)

    # write acn results to console and module logfile:
    runtime_output.print('ACN for CBR =  3 (ultra-low) is: ' + str(dict_acn_flex_parameter['PACN'][0]))
    runtime_output.print('ACN for CBR =  6 (low) is:       ' + str(dict_acn_flex_parameter['PACN'][1]))
    runtime_output.print('ACN for CBR = 10 (medium) is:    ' + str(dict_acn_flex_parameter['PACN'][2]))
    runtime_output.print('ACN for CBR = 15 (high) is:      ' + str(dict_acn_flex_parameter['PACN'][3]))

    runtime_output.debug('estimate_aircraft_classification_number successfully executed!')

    return dict_acn_flex_parameter, dict_constants
