# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing report functionalities for current calculation method."""
from yattag import Doc
from bs4 import BeautifulSoup
import os


def method_html_report(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """HTML report function.

    This function is responsible for creating reports.
    Note: Please ensure to format the data in the preparation part of the code (e.g., round floats)!

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict data_dict: Dictionary containing results of module execution
    :return: None
    """
    runtime_output.print(f'HTML report is generated and saved...')

    """ Preparation and data extraction."""
    # Extract tool and project name.
    tool_name = paths_and_names['tool_name']
    tool_version = paths_and_names['tool_version']
    report_title = (tool_name.replace('_', ' ')).capitalize()
    project_name = paths_and_names['name_of_project']
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))

    # Aircraft classification numbers.
    acn_data = [
        ("CBR", *data_dict['acn_estimation']['PCBR']),
        ("ACN", *data_dict['acn_estimation']['PACN'])
    ]
    acn_title = "Aircraft classification numbers"
    acn_column_titles = []  # table has no head -> no column titles necessary
    acn_column_alignment = ["l"]
    acn_column_alignment.extend(["r"] * len(data_dict['acn_estimation']['PCBR']))
    
    # Landing gear positions.
    positions_data = []
    for key, value in data_dict['landing_gear_positions'].items():
        formatted_key = key.replace('_', ' ').capitalize()
        if isinstance(value, (int, float)):
            formatted_value = f"{value:,.2f} m"
        else:
            formatted_value = value
        positions_data.append((formatted_key, formatted_value))
    
    positions_title = "Positions"
    positions_column_titles = ["Parameter", "Value"]
    positions_column_alignment = ["l", "r"]

    # Landing gear masses.
    nlg_masses = [
        (key.replace("_", " ").capitalize(), f"{value:,.2f} kg")
        for key, value in data_dict['landing_gear_masses']['nose_landing_gear_masses'].items()
    ]
    nlg_masses.sort(key=lambda x: "per strut" not in x[0])
    mlg_masses = [
        (key.replace("_", " ").capitalize(), f"{value:,.2f} kg")
        for key, value in data_dict['landing_gear_masses']['main_landing_gear_masses'].items()
    ]
    mlg_masses.sort(key=lambda x: "per strut" not in x[0])
    total_mass = [("Landing gear total mass", f"{data_dict['landing_gear_masses']['landing_gear_total_mass']:,.2f} kg")]
    masses = nlg_masses + mlg_masses + total_mass
    masses_title = "Masses"
    masses_column_titles = ["Component", "Mass"]
    masses_column_alignment = ["l", "r"]
    
    # Limitations.
    if data_dict['program_mode'] == 'mode_0':    
        limitations = [
            ("Tail tipping angle", f"{data_dict['landing_gear_limitations']['maximum_tail_tipping_angle']:,.2f} deg"),
            ("Turnover angle", f"{data_dict['landing_gear_limitations']['current_turn_over_angle']:,.2f} deg"),
            ("Wing tip clearance", f"{data_dict['landing_gear_limitations']['wing_tip_clearance_angles']:,.2f} deg"),
            ("Minimum forward position of nose gear",
            f"{data_dict['landing_gear_limitations']['minimum_forward_position_of_nose_gear']:,.2f} m"),
            ("Maximum aft position of nose gear",
            f"{data_dict['landing_gear_limitations']['maximum_aft_position_of_nose_gear']:,.2f} m"),
            ("Minimum forward position of main gear",
            f"{data_dict['landing_gear_limitations']['minimum_forward_position_of_main_gear']:,.2f} m")
        ]
    elif data_dict['program_mode'] == 'mode_1':
        limitations = [
            ("Tail tipping angle", f"{data_dict['landing_gear_limitations']['maximum_tail_tipping_angle']:,.2f} deg"),
            ("Turnover angle", "Not checked in use existing geometry mode."),
            ("Wing tip clearance", "Not checked in use existing geometry mode."),
            ("Minimum forward position of nose gear", "Not checked in use existing geometry mode."),
            ("Maximum aft position of nose gear", "Not checked in use existing geometry mode."),
            ("Minimum forward position of main gear", "Not checked in use existing geometry mode.")
        ]
    limitations_title = "Limitations"
    limitations_column_titles = ["Parameter", "Value"]
    limitations_column_alignment = ["l", "r"]

    # Nose gear assembly.
    if data_dict['program_mode'] == 'mode_0':
        nose_gear_data = [
            ("Nose gear tire design load",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['nose_gear_tire_design_load']:,.0f} lbs"),
            ("Number of nose gear struts",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_struts']:,.0f}"),
            ("Number of nose gear axis per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_axis_per_strut']:,.0f}"),
            ("Number of nose gear tires per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_tires_per_strut']:,.0f}"),
            ("Number of nose gear tires per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['nose_gear_strut_length']:,.2f} m"),
        ]
    elif data_dict['program_mode'] == 'mode_1':
        nose_gear_data = [
            ("Nose gear tire design load", "Not calculated in use existing geometry mode."),
            ("Number of nose gear struts",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_struts']:,.0f}"),
            ("Number of nose gear axis per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_axis_per_strut']:,.0f}"),
            ("Number of nose gear tires per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_tires_per_strut']:,.0f}"),
            ("Number of nose gear tires per strut",
            f"{data_dict['nose_gear_description']['nose_gear_arrangement']['nose_gear_strut_length']:,.2f} m"),
        ]     
    nose_gear_title = "Nose gear data"
    nose_gear_column_titles = ["Parameter", "Value"]
    nose_gear_column_alignment = ["l", "r"]

    # Nose gear tires.
    nose_gear_tires = []

    for key in data_dict['nose_gear_description']['nose_gear_tires'][next(iter(data_dict['nose_gear_description']['nose_gear_tires']))]:
        values = [
            value if key == "tire_size" or value == "not_available"
            else f"{round(float(value), 2):,.2f}"
            for value in (
                data_dict['nose_gear_description']['nose_gear_tires'][identifier][key]
                for identifier in data_dict['nose_gear_description']['nose_gear_tires']
            )
        ]
        nose_gear_tires.append((key.replace("_", " ").capitalize(), *values))

    nose_gear_tires_title = "Nose gear tire data"
    nose_gear_tires_column_titles = ["Parameter"] + list(data_dict['nose_gear_description']['nose_gear_tires'].keys())
    nose_gear_tires_column_alignment = ["l"]
    additional_alignments = len(data_dict['nose_gear_description']['nose_gear_tires'])
    nose_gear_tires_column_alignment.extend(["r"] * additional_alignments)

    # Main gear assembly.
    main_gear_data = []
    delete_list = ['local_tire_position_in_x_for_acn', 'local_tire_position_in_y_for_acn']

    for key, value in data_dict['main_gear_description']['main_gear_arrangement'].items():
        if key not in delete_list:
            formatted_key = key.replace('_', ' ').capitalize()
        else:
            continue
        
        if isinstance(value, (int, float)):
            if 'load' in key or any(word in key for word in ['position', 'track', 'length']):
                formatted_value = f"{value:,.2f}"
            else:
                formatted_value = f"{value:,.0f}"
        else:
            formatted_value = value

        if 'load' in key:
            formatted_value = f"{formatted_value} lbs"
        elif any(word in key for word in ['position', 'track', 'base', 'length']):
            formatted_value = f"{formatted_value} m"

        main_gear_data.append((formatted_key, formatted_value))
    main_gear_data = [item for item in main_gear_data if item not in delete_list]
    main_gear_title = "Main gear data"
    main_gear_column_titles = ["Parameter", "Value"]
    main_gear_column_alignment = ["l", "r"]

    # Main gear tires.
    main_gear_tires = []
    # for key in data_dict['main_gear_description']['main_gear_tires'][next(iter(data_dict['main_gear_description']['main_gear_tires']))]:
    #     values = [data_dict['main_gear_description']['main_gear_tires'][identifier][key] for identifier in data_dict['main_gear_description']['main_gear_tires']]
    #     main_gear_tires.append((key.replace("_", " ").capitalize(), *values))

    for key in data_dict['main_gear_description']['main_gear_tires'][next(iter(data_dict['main_gear_description']['main_gear_tires']))]:
        values = [
            value if key == "tire_size" or value == "not_available"
            else f"{round(float(value), 2):,.2f}"
            for value in (
                data_dict['main_gear_description']['main_gear_tires'][identifier][key]
                for identifier in data_dict['main_gear_description']['main_gear_tires']
            )
        ]
        main_gear_tires.append((key.replace("_", " ").capitalize(), *values))

    main_gear_tires_title = "Main gear tire data"
    main_gear_tires_column_titles = ["Parameter"] + list(data_dict['main_gear_description']['main_gear_tires'].keys())
    main_gear_tires_column_alignment = ["l"]
    additional_alignments = len(data_dict['main_gear_description']['main_gear_tires'])
    main_gear_tires_column_alignment.extend(["r"] * additional_alignments)
        
    """Generate report."""
    # Start Yattag document.
    doc, tag, text = Doc().tagtext()

    # Generate header (without creating a new <html>).
    with tag('html', lang="en"):
        # Generate report header (adds <head>).
        generate_report_header(doc, report_title)

        # Generate body.
        with tag('body'):
            with tag('div', klass="logo"):
                text("")
            with tag('div', klass="content"):
                with tag('h1'):
                    text(f"Report - {report_title} of {project_name}")
                    with tag('br'):  # Line break for formatting
                        pass
                    with tag('span'):
                        with tag('font', size="2"):
                            text(f"All outputs of the program '{tool_name}' were created with version {tool_version}")
                    
                with tag('div', klass="container"):
                    # Data on the left side.
                    with tag('div', klass="box data"):
                        
                        # General landing gear data.
                        with tag('h2'):
                            text("General landing gear data")
                        # Add table for ACN.
                        generate_table(doc, tag, text, acn_data, acn_title,
                                       acn_column_titles, acn_column_alignment, table_without_head=True)
                        # Add table for landing gear positions.
                        generate_table(doc, tag, text, positions_data, positions_title,
                                       positions_column_titles, positions_column_alignment)
                        # Add table for landing gear masses.
                        generate_table(doc, tag, text, masses, masses_title,
                                       masses_column_titles, masses_column_alignment, last_line_bold=True)
                        # Add table for landing gear limitations.
                        generate_table(doc, tag, text, limitations, limitations_title,
                                       limitations_column_titles, limitations_column_alignment)

                        # Nose landing gear data.
                        with tag('h2'):
                            text("Nose landing gear data")
                        # Add table for nose landing gear.
                        generate_table(doc, tag, text, nose_gear_data, nose_gear_title,
                                       nose_gear_column_titles, nose_gear_column_alignment)
                        # Add table for nose gear tires.
                        generate_table(doc, tag, text, nose_gear_tires, nose_gear_tires_title,
                                       nose_gear_tires_column_titles, nose_gear_tires_column_alignment)

                        # Main landing gear data.
                        with tag('h2'):
                            text("Main landing gear data")
                        # Add table for main landing gear.
                        generate_table(doc, tag, text, main_gear_data, main_gear_title,
                                       main_gear_column_titles, main_gear_column_alignment)
                        # Add table for main landing gear tires.
                        generate_table(doc, tag, text, main_gear_tires, main_gear_tires_title,
                                       main_gear_tires_column_titles, main_gear_tires_column_alignment)

                    # Plots on the right side.
                    # Ensure that plots are enabled before displaying them on report.
                    if plot_switch:
                        with tag('div', klass="box plot"):
                            with tag('h2'):
                                text("Plots")
                                # Arrangement.
                                with tag('h3'):
                                    plot_name = "Front view of landing gear arrangement"
                                    text(plot_name)
                                    src_name = "../plots/landing_gear_design_" + plot_name + ".svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass

    # Convert the document to HTML string.
    html_output = doc.getvalue()

    # Pretty HTML output with BeautifulSoup.
    soup = BeautifulSoup(html_output, 'html.parser')
    pretty_html = soup.prettify()

    """ Save report to current aircraft project directory. """
    # Save the report.
    project_directory = paths_and_names["project_directory"]
    report_directory = os.path.join(project_directory, 'reporting', 'report_html')
    os.makedirs(report_directory, exist_ok=True)
    output_path = os.path.join(report_directory, f"{tool_name}_report.html")

    # Write the HTML to the file.
    with open(output_path, "w", encoding="utf-8") as report:
        report.write(pretty_html)
    
    return


def generate_table(doc, tag, text, data, caption, column_titles, alignment=None, last_line_bold=False,
                   table_without_head=False):
    """Generate an HTML table.

    This function creates an HTML table from the provided data.
    Note: Options for the alignment types are "l" for left, "r" for right, and "m" for center. If not provided, columns
    are left-aligned by default.

    :param doc: Yattag Doc instance
    :param tag: Yattag tag function
    :param text: Yattag text function
    :param str caption: Table caption
    :param list column_titles: Titles of table columns
    :param list alignment: List of alignment types for ech column (defaults to None)
    :param bool last_line_bold: Flag if last table line should be printed bold
    :param bool table_without_head: Flag if table has no table head
    :param list data: List of tuples containing table data
    """
    with tag('table', klass="content-table"):
        with tag('caption'):
            text(caption)
        if table_without_head:
            pass
        else:
            with tag('thead'):
                with tag('tr'):
                    for col_num, col in enumerate(column_titles):
                        if alignment:
                            if alignment[col_num] == "m":
                                with tag('th', style="text-align: center;"):
                                    text(f"{col}")
                            elif alignment[col_num] == "r":
                                with tag('th', style="text-align: right;"):
                                    text(f"{col}")
                            elif alignment[col_num] == "l":
                                with tag('th', style="text-align: left;"):
                                    text(f"{col}")
                        else:                       
                            with tag('th'):
                                text(f"{col}")
        with tag('tbody'):
            row_number = 0
            for row in data:
                if last_line_bold:
                    is_last_row = (row_number == len(data) - 1)
                else:
                    is_last_row = False
                with tag('tr', style="font-weight: normal;" if is_last_row else ""):
                    column_number = 0
                    for col in row:
                        if col is None:
                            with tag('td'):
                                text('N.A.')
                        else:
                            if alignment is not None:
                                if alignment[column_number] == "m":
                                    with tag('td', style="text-align: center;"):
                                        text(col)
                                elif alignment[column_number] == "r":
                                    with tag('td', style="text-align: right;"):
                                        text(col)
                                elif alignment[column_number] == "l":
                                    with tag('td', style="text-align: left;"):
                                        text(col)
                                else:
                                    with tag('td'):
                                        text(col)
                            else:
                                with tag('td'):
                                    text(col)
                        column_number += 1
                row_number += 1
    
    return


def generate_report_header(doc, tool_name):
    """Generate HTML header.
    
    This function generates an HTML header with meta information, style sheets, and title.

    :param doc: Yattag Doc instance
    :param str tool_name: Name of the tool
    """
    with doc.tag('head'):
        doc.stag('meta', name="viewport", content="width=device-width, initial-scale=1.0")
        doc.stag('meta', charset="UTF-8")
        with doc.tag('title'):
             doc.text(tool_name)
        doc.stag('link', rel="stylesheet", href="style.css")
    return
