/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* List of references
    [1] Airbus customer services, Flight Operations Support & Line Assistance: Getting to grips with the cost index, Issue II, May 1998, p.7
*/
#ifndef MISSION_ANALYSIS_SRC_LIBS_MISSION_FILE_DATA_HANDLER_MISSION_FILE_DATA_HANDLER_H_
#define MISSION_ANALYSIS_SRC_LIBS_MISSION_FILE_DATA_HANDLER_MISSION_FILE_DATA_HANDLER_H_

#include <string>
#include <vector>

#include <aixml/node.h>
#include <moduleBasics/module.h>

/** \class MissionFileDataHandler
 *  \brief Exraction of data from the mission file
 */
class MissionFileDataHandler {
 public:
    /** \class OfftakeSchedule
     *  \brief Schedule for the offtakes during the mission
     */
    class OfftakeSchedule {
     public:
        double time; /**< Flight time [s] */
        double value; /**< Value of the offtake: power offtake [W], bleed offtake [kg/s] */
        /* Constructor and destructor */
        /** \brief Constructor for class OfftakeSchedule
         * \param a_time: a specific flight time [s]
         * \param a_Value: power offtake [W], bleed offtake [kg/s]
         */
        OfftakeSchedule(double a_time, double a_Value);
        /** \brief Destructor for class OfftakeSchedule
         */
        virtual ~OfftakeSchedule();
    };
    std::vector<OfftakeSchedule> bleed_schedule;       /**< Bleed schedule [s], [kg/s] */
    std::vector<OfftakeSchedule> shaft_power_schedule; /**< Shaft power schedule [s], [W] */
    /** \struct MissionDataSet
     *  \brief A struct that contains all the information which was exracted from the mission file
     */
    struct MissionDataSet {
        int mission_file_index;     /**< Index in mission file [-] */
        int mode;                   /**< Flight mode [-]: 1 = Takeoff; 2 = Accelerate; 3 = Climb; 4 = CruiseFL; 5 = CruiseMach; 6 = Descend; 7 = Land; 8 = Decelerate */
        std::string mode_name;      /**< Mode name */
        std::string rating; /**< Engine rating:'TO': acc. to EGTmax_TO; 'MaxCont': acc. to EGTmax_Cont; 'Climb': N1(MaxCont)*0.9; 'Cruise': N1(MaxCont)*0.85; 'Idle': N1(P4 min) */
        double derate;              /**< Relative derate to the selected rating [-] */
        std::string config;         /**< Aerodynamic configuration */
        double altitude;            /**< Final flight altitude [m] */
        double ROC;                 /**< Rate of climb [m/s] */
        double CAS;                 /**< End CAS [m/s] */
        double mach;                /**< Mach number [-] */
        bool mach_switch;           /**< Switch for using Mach */
        double glide_path;           /**< Glide angle [rad] */
        double rel_segment_length;  /**< Relative segment length in cruise */
        bool auto_FL_optimum;       /**< Switch to auto select optimum flight level */
        bool round_to_regular_FL;   /**< Switch for rounding to regular flight level */
        double auto_altitude_steps; /**< Altitude change for optimum altitude in cruise [ft] */
        double CI;                  /**< Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1] */
        bool CI_via_speed_switch;   /**< Switch for calculating velocity variably according to CI */
        double transfer_time;       /**< Transfer time [min] */
        double transfer_rate;       /**< Transfer rate [kg/min] */
        double rel_delta_fuel_mass; /**< Relative delta fuel mass [kg] */
        double delta_payload_mass;  /**< Delta payload mass [kg] */
        std::vector<OfftakeSchedule> bleed_schedule;       /**< Bleed schedule [s], [kg/s] */
        std::vector<OfftakeSchedule> shaft_power_schedule; /**< Shaft power schedule [s], [W] */
        double background_bleed;       /**< Background bleed [kg/s] */
        double background_shaft_power; /**< Background shaft power [W] */
        int energy_carrier_id;       /**< Energy carrier id for mission the segment */
    };
    std::vector<MissionDataSet> mission_data; /* Vector to gather the data from the mission XML steps */
    const node& mission_xml;         /* Reference to the mission XML */
    double Range;                    /**< Range [m] */
    double alternate_distance;       /**< Alternate distance [m] */
    double desired_cruise_speed;     /**< Planned cruise Mach number for fuel calculation [-] */
    double taxi_time_origin;         /**< Taxi time at departure airport [s] */
    double taxi_time_destination;    /**< Taxi time at destination airport [s] */
    bool taxiing_procedure;          /**< Switch for electrically driven taxiing (e.g. with fuel cell APU) */
    double engine_warmup_time;       /**< Time of engine warm-up with electric taxiing [s] */
    double terminal_operations_time; /**< Terminal time at destination airport for go around [s] */
    double air_to_air_fuel_mass;     /**< Air to air fuel mass [kg] */
    double payload_mass;             /**< Payload [kg] */
    unsigned int PAX;                /**< Passengers on board [-] */
    double cargo_mass;               /**< Air cargo on board [kg] */

    /* Constructor and destructor */
    /** \brief Constructor for class FlightConditions
     * \param path_to_mission_xml: reference to the mission XML
     */
    explicit MissionFileDataHandler(const std::string& path_to_mission_xml);
    /** \brief Destructor for class FlightConditions
     */
    virtual ~MissionFileDataHandler();

    /* Member methods */
    /** \brief Function reads the departure block in current mission_xml file
     *  \param max_operating_speed: reference to the maximum operating speed [m/s]
     */
    void read_departure_steps(const double& max_operating_speed, const node& energy_provider);
    /** \brief Function reads the flight block in current mission_xml file
     *  \param average_shaft_bleed: reference to the average shaft power [W]
     *  \param average_cruise_bleed: reference to the average cruise bleed [kg/s]
     *  \param auto_climb_switch: Switch to activate automatic climb
     *  \param step_height_cruise: reference to the step height for automatic climb [ft]
     *  \param max_operating_mach: reference to the maximum operating mach number [-]
     *  \param max_operating_speed: reference to the maximum operating speed [m/s]
     */
    void read_flight_steps(double* average_shaft_bleed, double* average_cruise_bleed, bool* auto_climb_switch, double* step_height_cruise,
                            const double& max_operating_mach, const double& max_operating_speed, const node& energy_provider);
    /** \brief Function reads the approach block in current mission_xml file
     *  \param max_operating_speed: reference to the maximum operating speed [m/s]
     */
    void read_approach_steps(const double& max_operating_speed, const node& energy_provider);
};

#endif // MISSION_ANALYSIS_SRC_LIBS_MISSION_FILE_DATA_HANDLER_MISSION_FILE_DATA_HANDLER_H_
