/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "multi_engine_propulsion.h"
#include <algorithm>

/* Initialization of attributes of the MultiEnginePropulsion object with the constructor */
MultiEnginePropulsion::MultiEnginePropulsion(const std::shared_ptr<RuntimeIO>& rtIO) {
    // Set up needed nodes
    const node& propulsor_requirements = rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion");
    const node& propulsor_design = rtIO->acxml.at("component_design/propulsion/specific");
    const node& tank_definition = rtIO->acxml.at("requirements_and_specifications/design_specification/configuration/tank_definition");
    const node& tank_design = rtIO->acxml.at("component_design/tank/specific");

    /* Check if all tanks and engines need the same energy carriers and if they were defined */
    std::vector<int> propulsor_energy_carriers(propulsor_requirements.getIntVector("propulsor/energy_carrier_ID/value"));
    std::vector<int> tank_energy_carriers(tank_definition.getIntVector("tank/energy_carrier_ID/value"));
    std::vector<int> available_energy_carriers;
    std::transform(rtIO->aircraft_energy_carriers().begin(), rtIO->aircraft_energy_carriers().end(),
                    std::back_inserter(available_energy_carriers), [](const std::pair<const int, std::tuple<std::string, double>>& energy_carrier)
                    { return energy_carrier.first; });
    if (!vector_check(propulsor_energy_carriers, tank_energy_carriers)) {
        throwError(__FILE__, __func__, __LINE__,
           "Check your energy carriers: Given energy carrier IDs for propulsion system and tanks do not match!");
    }
    if (!vector_check(propulsor_energy_carriers, available_energy_carriers)) {
        throwError(__FILE__, __func__, __LINE__,
            "Check your energy carriers: Propulsion system and tanks contain energy carrier IDs which were not defined or you defined energy carriers that are not used!");
    }

    /* Get all propulsors  */
    for (uint16_t i(0); i < propulsor_design.getVector("propulsion").size(); i++) {
        // get the energy_carrier_id of the propulsor
        int id_tmp = EndnodeReadOnly<int>("/propulsor@" + num2Str(i) + "/energy_carrier_ID").read(propulsor_requirements).value();

        // get all characteristic for its engine
        double scale_tmp(EndnodeReadOnly<double>("/propulsion@" + num2Str(i) + "/engine/scale_factor").read(propulsor_design).value());
        std::string name_tmp = EndnodeReadOnly<std::string>("/propulsion@" + num2Str(i) + "/engine/model").read(propulsor_design).value();
        double thrust_share_tmp(EndnodeReadOnly<double>("/propulsor@" + num2Str(i) + "/thrust_share").read(propulsor_requirements).value());

        // Check if the energy_carrier_id is already used. If not, setup new propulsion system for it
        if (this->propulsors.find(id_tmp) == this->propulsors.end()) {
            this->energy_map.emplace(id_tmp,
                EnergyConsumption(EnergyCarrier(rtIO->get_fuel_type(id_tmp), rtIO->get_fuel_density(id_tmp))));
            Propulsor propulsor_tmp(name_tmp, Engine(rtIO->getEngineDataDir() + "/" + name_tmp, scale_tmp), i, thrust_share_tmp);
            this->propulsors[id_tmp].push_back(propulsor_tmp);
        } else { // If it already exists, add a new one or update ID vector for equal engines
            for (uint16_t j(0); j < this->propulsors[id_tmp].size(); j++) {
                if (this->propulsors[id_tmp][j].name == name_tmp // check for same name and scale factor
                        && accuracyCheck(this->propulsors[id_tmp][j].engine.get_scale_factor(), scale_tmp, ACCURACY_LOW)) {
                    this->propulsors[id_tmp][j].engine_id.push_back(i); // Add the ID of the same engine to avoid double effort calculating engines
                } else {
                    Propulsor propulsor_tmp(name_tmp, Engine(rtIO->getEngineDataDir() + "/" + name_tmp, scale_tmp), i, thrust_share_tmp);
                    this->propulsors[id_tmp].push_back(propulsor_tmp);
                }
            }
        }
    }

    /* Get all tanks and gather them with corresponding energy carrier IDs */
    // Read tank IDs and capacities from the aircraft XML
    std::map<int, double> total_capacity_per_id;
    for (int tank_id(0); tank_id < tank_design.getVector("tank").size(); tank_id++) {
        //Only take tanks into account that are used for the mission
        if (EndnodeReadOnly<bool>("energy_capacity_required_for_mission").read(tank_design.at("tank@" + num2Str(tank_id))).value()) {
            int energy_carrier_id(EndnodeReadOnly<int>("energy_carrier_ID").read(tank_definition.at("tank@" + num2Str(tank_id))).value());
            double tank_capacity(EndnodeReadOnly<double>("maximum_energy_capacity").read(tank_design.at("tank@" + num2Str(tank_id))).value());
            Tank tmp_tank(tank_id, tank_capacity);
            this->energy_map.at(energy_carrier_id).tanks.push_back(tmp_tank);
            total_capacity_per_id[energy_carrier_id] += tank_capacity; // sum capacity of this energy carrier
        }
    }
    // Get the relative maximum capacity of each tank in comparison to other tanks using the same energy carrier ID
    for (const auto& [energy_carrier_id, energy_consumtion] : this->energy_map) {
        for (uint16_t i(0); i < energy_consumtion.tanks.size(); i++) {
            this->energy_map.at(energy_carrier_id).tanks[i].relative_capacity =
                energy_consumtion.tanks[i].capacity/total_capacity_per_id[energy_carrier_id];
        }
    }
}

/* Destructor of the flight conditions object */
MultiEnginePropulsion::~MultiEnginePropulsion() {
}

size_t MultiEnginePropulsion::unique_engines_per_id(const int& energy_carrier_id) {
    return this->propulsors[energy_carrier_id].size();
}

uint16_t MultiEnginePropulsion::get_number_of_engines(const int& energy_carrier_id) {
    const std::vector<MultiEnginePropulsion::Propulsor>& propulsor_vec = this->propulsors.at(energy_carrier_id);
    return std::accumulate(
        propulsor_vec.begin(), propulsor_vec.end(), 0, [](uint16_t sum, const MultiEnginePropulsion::Propulsor& propulsor) {
            return sum + propulsor.engine_id.size();});
}

std::vector<OperatingPoint> MultiEnginePropulsion::get_operating_points(const int& energy_carrier_id) {
    std::vector<OperatingPoint> operating_points;
    for (uint16_t i(0); i < this->propulsors[energy_carrier_id].size(); i++) {
        operating_points.push_back(this->propulsors[energy_carrier_id][i].engine.get_operating_point());
    }
    return operating_points;
}

std::vector<MultiEnginePropulsion::Propulsor> MultiEnginePropulsion::get_propulsors(const int& energy_carrier_id) {
    std::vector<Propulsor> propulsors_vector;
    std::copy(this->propulsors[energy_carrier_id].begin(), this->propulsors[energy_carrier_id].end(), std::back_inserter(propulsors_vector));
    return propulsors_vector;
}

void MultiEnginePropulsion::set_operating_points(const int& energy_carrier_id, std::vector<OperatingPoint> operating_points) {
    for (const auto& [propulsor_energy_carrier_id, propulsor] : this->propulsors) {
        for (uint16_t i(0); i < propulsor.size(); i++) {
            if (propulsor_energy_carrier_id != energy_carrier_id) {
                OperatingPoint n0_operating_point(operating_points[i]); // Copy operating point since Mach number and altitude are euqal for all
                n0_operating_point.N = 0.;                              // Set N1 for engines which are not running to 0
                this->propulsors[energy_carrier_id][i].engine.set_operating_point(n0_operating_point);
            } else {
                this->propulsors[energy_carrier_id][i].engine.set_operating_point(operating_points[i]);
            }
        }
    }
}

double MultiEnginePropulsion::get_aircraft_fuelflow(const int& energy_carrier_id) {
    double total_fuel_flow(0);
    // Loop through all propulsors running on that energy carrier
    for (uint16_t i(0); i < this->propulsors[energy_carrier_id].size(); i++) {
        // Trick: Propulsors with same engines have multiple IDs -> size() scales overall fuelflow acordingly
        total_fuel_flow += this->propulsors[energy_carrier_id][i].engine.get_fuelflow() * this->propulsors[energy_carrier_id][i].engine_id.size();
    }
    return total_fuel_flow;
}

double MultiEnginePropulsion::get_aircraft_energyflow(const int& energy_carrier_id) {
    return this->get_aircraft_fuelflow(energy_carrier_id) * this->energy_map.at(energy_carrier_id).energy_carrier.gravimetric_energy_density;
}

double MultiEnginePropulsion::get_aircraft_energyflow(const int& energy_carrier_id, const double& aircraft_fuelflow) {
    return aircraft_fuelflow * this->energy_map.at(energy_carrier_id).energy_carrier.gravimetric_energy_density;
}

double MultiEnginePropulsion::get_thrust_aircraft(const int& energy_carrier_id) {
    double total_thrust(0);
    for (uint16_t i(0); i < this->propulsors[energy_carrier_id].size(); i++) {
        // Trick: Propulsors with same engines have multiple IDs -> size() scales overall thrust acordingly
        total_thrust += this->propulsors[energy_carrier_id][i].engine.get_thrust() * this->propulsors[energy_carrier_id][0].engine_id.size();
    }
    return total_thrust;
}

void MultiEnginePropulsion::calculate_N1_with_penalties(const int& energy_carrier_id, const double& altitude,
                                             const double& mach_number, const atmosphere& atm,
                                             const double& derate, const std::string& thrust_rating,
                                             double bleed_air_offtake, double shaft_power_offtake) {
    /* Loop through all engines using the energy carrier ID and set it to its N1 */
    for (uint16_t i(0); i < this->propulsors[energy_carrier_id].size(); i++) {
        // Assume power and bleed offtakes are equal to thrust share as long as nothing is specified
        double shaft_power_offtake_share = this->propulsors[energy_carrier_id][i].thrust_share*shaft_power_offtake;
        double bleed_air_offtake_share = this->propulsors[energy_carrier_id][i].thrust_share*bleed_air_offtake;
        this->propulsors[energy_carrier_id][i].engine.calculate_N1_with_penalties(
                                                        altitude, mach_number, atm, derate,
                                                        thrust_rating, bleed_air_offtake_share, shaft_power_offtake_share);
    }
}

void MultiEnginePropulsion::calculate_N1_with_thrustlimit(const int& energy_carrier_id, const double& altitude,
                                             const double& mach_number, const atmosphere& atm,
                                             const double& derate, const std::string& thrust_rating,
                                             double bleed_air_offtake, double shaft_power_offtake, double thrust_limit) {
    /* Loop through all engines using the energy carrier ID and set it to its N1 using a thrust limit */
    for (uint16_t i(0); i < this->propulsors[energy_carrier_id].size(); i++) {
        // Assume power and bleed offtakes are equal to thrust share as long as nothing is specified
        double thrust_limit_share = this->propulsors[energy_carrier_id][i].thrust_share*thrust_limit;
        double shaft_power_offtake_share = this->propulsors[energy_carrier_id][i].thrust_share*shaft_power_offtake;
        double bleed_air_offtake_share = this->propulsors[energy_carrier_id][i].thrust_share*bleed_air_offtake;
        this->propulsors[energy_carrier_id][i].engine.calculate_N1_with_thrustlimit(
                                                        altitude, mach_number, atm, derate, thrust_rating,
                                                        bleed_air_offtake_share, shaft_power_offtake_share, thrust_limit_share);
    }
}

std::map<int, double> MultiEnginePropulsion::get_consumed_energy_map() {
    std::map<int, double> consumed_energy_map;
    for (const auto& [energy_carrier_id, map] : this->energy_map) {
        consumed_energy_map[energy_carrier_id] = get_consumed_energy(energy_carrier_id);
    }
    return consumed_energy_map;
}

double MultiEnginePropulsion::get_consumed_energy(const int& energy_carrier_id) {
    double total_consumed_energy = std::accumulate(this->energy_map.at(energy_carrier_id).tanks.begin(), this->energy_map.at(energy_carrier_id).tanks.end(), 0.0,
            [](double sum, const Tank& tank) {return sum + tank.consumed_energy;});
    return total_consumed_energy;
}

std::map<int, double> MultiEnginePropulsion::get_energyflow_map() {
    std::map<int, double> energyflow_map;
    for (const auto& [energy_carrier_id, map] : this->energy_map) {
        energyflow_map[energy_carrier_id] =
            this->get_aircraft_fuelflow(energy_carrier_id) * map.energy_carrier.gravimetric_energy_density;
    }
    return energyflow_map;
}

void MultiEnginePropulsion::drain_tanks(const int& energy_carrier_id, const double& delta_t, const double& fuel_mass) {
    // TODO(Gerrit): Only share on overall energy is defueled. Other methods could be implemented
    if (std::isnan(fuel_mass)) {
        for (uint16_t i(0); i < this->energy_map.at(energy_carrier_id).tanks.size(); i++) {
            this->energy_map.at(energy_carrier_id).tanks[i].consumed_energy -=  delta_t *
                get_aircraft_energyflow(energy_carrier_id) * this->energy_map.at(energy_carrier_id).tanks[i].relative_capacity;
        }
    } else { // If a given fuel mass is taken away, just use that one
        for (uint16_t i(0); i < this->energy_map.at(energy_carrier_id).tanks.size(); i++) {
            this->energy_map.at(energy_carrier_id).tanks[i].consumed_energy -=  fuel_mass *
                this->energy_map.at(energy_carrier_id).energy_carrier.gravimetric_energy_density *
                this->energy_map.at(energy_carrier_id).tanks[i].relative_capacity;
        }
    }
}

double MultiEnginePropulsion::get_consumed_fuel_mass() {
    std::map<int, double> consumed_energy_map(this->get_consumed_energy_map());
    double total_fuel_mass = std::accumulate(this->energy_map.begin(), this->energy_map.end(), 0.0,
        [&consumed_energy_map](double sum, const std::pair<const int, MultiEnginePropulsion::EnergyConsumption>& energy_map_tmp) {
            return sum + consumed_energy_map.at(energy_map_tmp.first) / energy_map_tmp.second.energy_carrier.gravimetric_energy_density;
    });
    return total_fuel_mass;
}

double MultiEnginePropulsion::get_energy_carrier_capacity(const int& energy_carrier_id) {
    double total_capacity = std::accumulate(this->energy_map.at(energy_carrier_id).tanks.begin(), this->energy_map.at(energy_carrier_id).tanks.end(), 0.0,
            [](double sum, const Tank& tank) {return sum + tank.capacity;});
    return total_capacity;
}

/* Initialization of attributes of the Propulsor object with the constructor */
MultiEnginePropulsion::Propulsor::Propulsor(const std::string& a_name, Engine an_engine, const uint16_t& an_id, const double& a_thrust_share) :
    name(a_name),
    engine(an_engine),
    engine_id({an_id}),
    thrust_share(a_thrust_share) {
}

/* Destructor of the Propulsor object */
MultiEnginePropulsion::Propulsor::~Propulsor() {
}

/* Initialization of attributes of the Tank object with the constructor */
MultiEnginePropulsion::Tank::Tank(uint16_t a_tank_id, double a_capacity) :
    tank_id(a_tank_id),
    capacity(a_capacity),
    relative_capacity(0.),
    consumed_energy(0.) {
}

/* Destructor of the Tank object */
MultiEnginePropulsion::Tank::~Tank() {
}

/* Initialization of attributes of the EnergyConsumption object with the constructor */
MultiEnginePropulsion::EnergyConsumption::EnergyConsumption(const EnergyCarrier& an_energy_carrier) :
    energy_carrier(an_energy_carrier),
    tanks() {
}

/* Destructor of the EnergyCarrier object */
MultiEnginePropulsion::EnergyConsumption::~EnergyConsumption() {
}

bool MultiEnginePropulsion::vector_check(const std::vector<int>& v1, const std::vector<int>& v2) {
    std::set<int> set1(v1.begin(), v1.end());
    std::set<int> set2(v2.begin(), v2.end());

    return set1 == set2;
}
