/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include "performance_assessment_factory.h"
#include "performance_assessment_output.h"  // Resolving circular dependency

namespace taw {

performance_assessment_factory::performance_assessment_factory(const std::shared_ptr<RuntimeIO>& rtIOin)
    :
    rtIO(rtIOin),
    mySettings(rtIOin),
    mySettingsPt(std::make_shared<performance_assessment_settings>(mySettings)),
    myAcft(rtIO, mySettingsPt),
    myAcftPt(std::make_shared<aircraft>(myAcft)) {
        // ctor
    }

void performance_assessment_factory::initialize() {
    /* Get desired fidelity level of methods*/
    fidelityLevelFuelPlanning = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/fuel_planning/fidelity_level", "low").read(rtIO->moduleConfig).value();
    fidelityLevelOptimizeMission = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/mission/fidelity_level", "low").read(rtIO->moduleConfig).value();
    fidelityLevelPayloadRangeDiagram = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/payload_range_diagram/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelEnginePerformance = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/engine_performance/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelFlightEnvelopePerformance = EndnodeReadOnly<std::string>(
        "/module_configuration_file/program_settings/performance_checks/flight_envelope_performance/fidelity_level", "low").read(rtIO->moduleConfig).value();
    fidelityLevelClimbPerformance = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/climb_performance/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelTakeOffPerformance = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/takeoff_performance/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelLandingPerformance = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/landing_performance/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelVnDiagram = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/vn_diagram/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();
    fidelityLevelSARPerformance = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/performance_checks/sar_performance/fidelity_level", "low")
        .read(rtIO->moduleConfig).value();

    /* Initialization of performance_assessment objects */

    /* PayloadRange */
    if (fidelityLevelPayloadRangeDiagram == "high") {
        // ...
    } else if (fidelityLevelPayloadRangeDiagram == "low") {
        abstractPayloadRangePt = std::make_shared<low::payloadRange>(&myAcft, mySettings);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'PayloadRange' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* EnginePerformance */
    if (fidelityLevelEnginePerformance == "high") {
        // ...
    } else if (fidelityLevelEnginePerformance == "low") {
        abstractEnginePerformancePt = std::make_shared<low::enginePerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'EnginePerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* ClimbPerformance */
    if (fidelityLevelClimbPerformance == "high") {
        // ...
    } else if (fidelityLevelClimbPerformance == "low") {
        abstractCeilingPerformancePt = std::make_shared<low::ceilingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'ClimbPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Flight Envelope */
    if (fidelityLevelFlightEnvelopePerformance == "high") {
        // ...
    } else if (fidelityLevelFlightEnvelopePerformance == "low") {
        abstractFlightEnvelopesPt = std::make_shared<low::flightEnvelopes>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'FlightEnvelope' selected! Exiting"<< std::endl;
        exit(1);
    }

    /* SARPerformance */
    if (fidelityLevelSARPerformance == "high") {
        // ...
    } else if (fidelityLevelSARPerformance == "low") {
        abstractSARPerformancePt = std::make_shared<low::SARPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'SARPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Take-Off performance */
    if (fidelityLevelTakeOffPerformance == "high") {
        // ...
    } else if (fidelityLevelTakeOffPerformance == "low") {
        abstractTakeOffPerformancePt = std::make_shared<low::takeOffPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'takeOffPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Landing Performance */
    if (fidelityLevelLandingPerformance == "high") {
        // ...
    } else if (fidelityLevelLandingPerformance == "low") {
        abstractLandingPerformancePt = std::make_shared<low::landingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'landingPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Vn-Diagram */
    if (fidelityLevelVnDiagram == "high") {
        // ...
    } else if (fidelityLevelVnDiagram == "low") {
        abstractVnDiagramPt = std::make_shared<low::VnDiagram>(myAcftPt, mySettingsPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'VnDiagram' selected! Exiting"<< std::endl;
        exit(1);
    }
}  // initialize()


void performance_assessment_factory::run() {
/* Run the selected modules */

/* PayloadRange */
try {
    myRuntimeInfo->out << "      ... Payload Range" << std::endl;
    abstractPayloadRangePt->calcPayloadRange();
    mySettingsPt->check_sucessfully_executed.doPayloadRange = true;
} catch (const int& errorCode) {
    myRuntimeInfo->err << "Error in payload range calculation. See logLevel=2 (Info) for more details." << std::endl;
    mySettingsPt->check_sucessfully_executed.doPayloadRange = false;
}

/* EnginePerformance */
if (false) {  // Engine performance is teporarily disabled
    try {
        myRuntimeInfo->out << "  -> Engine Performance" << std::endl;
        abstractEnginePerformancePt->doEngineCalculation(mySettingsPt->enginePerformance.Rating);
        mySettingsPt->check_sucessfully_executed.doEnginePerformance = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doEnginePerformance = false;
        myRuntimeInfo->out << "Error in engine performance evaluation. See logLevel=2 (Info) for more details." << std::endl;
    }
}

/* Climb Performance */
try {
    myRuntimeInfo->out << "  -> Ceiling Performance" << std::endl;
    abstractCeilingPerformancePt->doCeilingCalculation();
    mySettingsPt->check_sucessfully_executed.doClimbPerformance = true;
    try {
    myRuntimeInfo->out << "  -> Flight Envelopes" << std::endl;
        if (fabs(myAcftPt->MDive - 0) < ACCURACY_HIGH || fabs(myAcftPt->VDive - 0.) < ACCURACY_HIGH || mySettingsPt->overwrite_initial_values) {
            abstractFlightEnvelopesPt->doDiveSpeedLimits();
        }
        abstractFlightEnvelopesPt->doFlightEnvelope(abstractCeilingPerformancePt);
        mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance = true;
        if (false) {  // SAR Performance is temporarily disabled
            myRuntimeInfo->out << "  -> SAR Performance" << std::endl;
            abstractSARPerformancePt->doSARCalculation(abstractFlightEnvelopesPt);
            mySettingsPt->check_sucessfully_executed.doSARCalculation = true;
        }
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance = false;
        myRuntimeInfo->err << "Error in flight envelope calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
    
} catch (const int& errorCode) {
    mySettingsPt->check_sucessfully_executed.doClimbPerformance = false;  // Prevents generation of plots with nonexisting data
    myRuntimeInfo->out << "Error in climb performance calculation. See logLevel=2 (Info) for more details." << std::endl;
}


/* Takeoff Performance*/
try {
    myRuntimeInfo->out << "  -> Takeoff Performance" << std::endl;
    abstractTakeOffPerformancePt->doTakeOffPerformance(mySettingsPt->myField.runwaySlope,
            myAcftPt->frictionCoefficient,
            myAcftPt->brakingCoefficient,
            mySettingsPt->myField.headWind);
            mySettingsPt->check_sucessfully_executed.doTOPerformance = true;
} catch (const int& errorCode) {
    mySettingsPt->check_sucessfully_executed.doTOPerformance = false;  // Prevents generation of plots with nonexisting data
    myRuntimeInfo->out << "Error in takeoff performance calculation. See logLevel=2 (Info) for more details." << std::endl;
}

if (false) {  // Balanced Field Length is teporarily disabled
    try {
        // Calculation of Balance Field Length
        myRuntimeInfo->out << "  -> Calculation of BLFL" << std::endl;
        abstractTakeOffPerformancePt->calcBLFL(mySettingsPt->myField.runwaySlope,
                                        myAcftPt->frictionCoefficient,
                                        myAcftPt->brakingCoefficient,
                                        mySettingsPt->myField.headWind);
        mySettingsPt->check_sucessfully_executed.doBLFLPerformance = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doBLFLPerformance = false;
        myRuntimeInfo->out << "Error in balanced field length calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
}



/* Landing Performance */
try {
    myRuntimeInfo->out << "  -> Landing Performance" << std::endl;
    abstractLandingPerformancePt->doLandingPerformance(mySettingsPt->myField.runwaySlope,
    myAcftPt->brakingCoefficient,
    mySettingsPt->myField.headWind);
    mySettingsPt->check_sucessfully_executed.doLandingPerformance = true;
} catch (const int& errorCode) {
    mySettingsPt->check_sucessfully_executed.doLandingPerformance = false;  // Prevents generation of plots with nonexisting data
    myRuntimeInfo->out << "Error in landing performance calculation. See logLevel=2 (Info) for more details." << std::endl;
}

/* Vn-Diagram */
if (false) {  // Vn-Diagram is teporarily disabled
    try {
        myRuntimeInfo->out << "  -> V-n Diagram" << std::endl;
        abstractVnDiagramPt->doVnDiagram();
        mySettingsPt->check_sucessfully_executed.doVnDiagram = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doVnDiagram = false;  // Prevents generation of plots with nonexisting data
        myRuntimeInfo->out << "Error in Vn diagram calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
}
}  // end of run()

void performance_assessment_factory::update() {
  // update
}

void performance_assessment_factory::report() {
    performance_assessment_output theOutput(
        abstractPayloadRangePt,
        abstractEnginePerformancePt,
        abstractCeilingPerformancePt,
        abstractFlightEnvelopesPt,
        abstractSARPerformancePt,
        abstractTakeOffPerformancePt,
        abstractLandingPerformancePt,
        abstractVnDiagramPt,
        myAcftPt);

    myRuntimeInfo->out << "  -> Generating output" << std::endl;
    theOutput.writeXmlOutput();
    theOutput.generatePlotData();
    theOutput.generatePlots();
    theOutput.generateHtmlBody();
}

void performance_assessment_factory::save() {
  // save data if necessary
}

}  // namespace taw
