/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* === Includes === */
#include <gtest/gtest.h>

/* Unit Under Test */
#include "engine_design/rubber/rubber_design.h"

/* === Fixtures === */
class TemporaryEngineDesign : public ::testing::Test
{
  protected:
    void SetUp() override
    {
        /* Create a temporary directory */
        temp_dir_ = std::filesystem::temp_directory_path() / "test_rubber_turbofan";
        std::filesystem::create_directory(temp_dir_);
    }

    void TearDown() override
    {
        /* Remove the temporary directory */
        std::filesystem::remove_all(temp_dir_);
    }

    std::filesystem::path temp_dir_;
};

/* === Tests === */
/**
 * @brief Test designing a turbofan engine with kerosene as fuel.
 */
TEST(EngineDesignRubber, DesignTurbofanKeroseneManuallySelected)
{
    /* Set the manual selection to the test engine */
    std::filesystem::path config_path{CMAKE_TEST_CONFIG_FILE};
    auto config = aixml::openDocument(config_path);
    config->at("module_configuration_file/program_settings/propulsion@Default/Rubber/engine_model/value") = "test-engine";

    /* Set the engine database path */
    config->at("module_configuration_file/program_settings/path_engine_database/value") = std::string{CMAKE_TEST_STUBS_DIR} + "/engines";

    /* Create the engine designer and initialize */
    design::Rubber designer{config, {""}, {}}; // No engine directory -> Use database entry
    designer.initialize();

    /* Design the engine which needs no scaling */
    Turbofan<EnergyCarrier::Kerosene> engine{};
    engine.set_required_thrust(110.31e3);
    designer(engine);

    /* Check the result */
    EXPECT_EQ(engine.model(), "test-engine");
    EXPECT_NEAR(engine.scale(), 1.0, 1e-6);
    EXPECT_NEAR(engine.dimension().height, 2.118, 1e-3);
    EXPECT_NEAR(engine.dimension().width, 1.829, 1e-3);
    EXPECT_NEAR(engine.dimension().length, 2.508, 1e-3);
    EXPECT_NEAR(engine.fan().diameter, 1.613, 1e-3);
    EXPECT_NEAR(engine.pointmass(Component::Engine).mass, 2404.0, 1e-3);
    EXPECT_NEAR(engine.bucket_point().thrust, 35090.0, 1e-3);
    EXPECT_NEAR(engine.bucket_point().tsfc, 0.0, 1e-3);

    /* Reduce the design thrust so a scale of 2 is necessary */
    engine.set_required_thrust(110.31e3 * 2);
    designer(engine);
    EXPECT_NEAR(engine.scale(), 2.0, 1e-6);
    EXPECT_NEAR(engine.dimension().height, 2.995, 1e-3);
    EXPECT_NEAR(engine.dimension().width, 2.586, 1e-3);
    EXPECT_NEAR(engine.dimension().length, 3.309, 1e-3);
    EXPECT_NEAR(engine.fan().diameter, 2.281, 1e-3);

}

/**
 * @brief Test designing a turbofan engine with kerosene as fuel.
 */
TEST(EngineDesignRubber, DesignTurbofanKeroseneAutomaticallySelected)
{
    /* Set the manual selection to the test engine */
    std::filesystem::path config_path{CMAKE_TEST_CONFIG_FILE};
    auto config = aixml::openDocument(config_path);
    config->at("module_configuration_file/program_settings/propulsion@Default/Rubber").deleteChild("engine_model");

    /* Set the engine database path */
    config->at("module_configuration_file/program_settings/path_engine_database/value") = std::string{CMAKE_TEST_STUBS_DIR} + "engines";

    /* Create the engine designer and initialize */
    design::Rubber designer{config, {""}, {}}; // No engine directory -> Use database entry
    designer.initialize();

    /* Design the engine which needs no scaling */
    Turbofan<EnergyCarrier::Kerosene> engine{};
    engine.set_required_thrust(110.31e3);
    designer(engine);

    /* Check the result */
    EXPECT_EQ(engine.model(), "test-engine");
}

/**
 * @brief Test that the save method copies non existing engines from the database.
 */
TEST_F(TemporaryEngineDesign, SaveTurbofanKerosene)
{
    /* Set the manual selection to the test engine */
    std::filesystem::path config_path{CMAKE_TEST_CONFIG_FILE};
    auto config = aixml::openDocument(config_path);
    config->at("module_configuration_file/program_settings/propulsion@Default/Rubber/engine_model/value") = "test-engine";

    /* Set the engine database path */
    config->at("module_configuration_file/program_settings/path_engine_database/value") = std::string{CMAKE_TEST_STUBS_DIR} + "/engines";

    /* Create the engine designer and initialize */
    design::Rubber designer{config, {temp_dir_}, {}}; // No engine directory -> Use database entry
    designer.initialize();
    Turbofan<EnergyCarrier::Kerosene> engine{};
    engine.set_required_thrust(110.31e3);
    designer(engine);
    designer.save();

    /* Check whether the engine directory is copied */
    auto result = temp_dir_ / "test-engine";
    ASSERT_TRUE(std::filesystem::exists(result));

    /* Calling save again should not add another folder */
    designer.save();
    auto n_folers = std::count_if(
        std::filesystem::directory_iterator(temp_dir_),
        std::filesystem::directory_iterator(),
        [](const auto &entry) { return entry.is_directory(); });
    EXPECT_EQ(n_folers, 1);
}

/**
 * @brief Test changing the technology factor of the engine efficiency.
 */
TEST(EngineDesignRubber, ChangeTechnologyFactor)
{
    /* Set the manual selection to the test engine */
    std::filesystem::path config_path{CMAKE_TEST_CONFIG_FILE};
    auto config = aixml::openDocument(config_path);
    config->at("module_configuration_file/program_settings/propulsion@Default/Rubber/engine_model/value") = "test-engine";

    /* Set the engine database path */
    config->at("module_configuration_file/program_settings/path_engine_database/value") = std::string{CMAKE_TEST_STUBS_DIR} + "/engines";

    /* Modify the technology factor */
    config->at("module_configuration_file/program_settings/technology_factors/engine_efficiency/value") = 0.9;

    /* Create the engine designer and initialize */
    design::Rubber designer{config, {""}, {}}; // No engine directory -> Use database entry
    EXPECT_THROW(designer.initialize(), std::runtime_error);
}
