/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEM_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEM_H_

#include <string>
#include <vector>
#include <memory>
#include <aixml/node.h>
#include <aixml/endnode.h>
#include <moduleBasics/runtimeIO.h>

#include "powerProfile.h"
#include <svl/SVL.h>
#include <aircraftGeometry2/processing/measure.h>

#define UNUSED(x) (void)(x) //can be used for unused variable in override functions of specific aircraft systems to avoid linter warnings

class systemsIOData;

/** \brief general class of aircraft system. Inherits to all specific systems (e.g. conventionalATA25, etc.)
 */
class aircraftSystem {
 protected:
    std::shared_ptr<systemsIOData> data_; /*<-- IOData object*/
    const RuntimeIO& rtIO_;

 public:
    /** \brief initializes the power profile for power demand during the mission segments
     *
     * \return void
     *
     */
    void initializePowerProfile();

    /** \brief calculates the maximum power demand during the mission
     *
     * \return void
     *
     */
    void calculateMaxMissionPower();

    /** \brief calculates the maximum power profile sum of electrical systems demand during the mission
     *
     * \return void
     *
     */
    double getElectricPower();

    /** \brief calculates the maximum  Hydraulic power profile sum demand during the mission
     *
     * \return void
     *
     */
    double getHydraulicPower();

    /** \brief updates system mass in acxml
     * \param
     * \return
    */
    void updateMass(std::string subPath, double value, std::string ATA);

    std::string Name;/**< name of the system (e.g. conventionalFurnishing) */
    std::string ATA;/**< ATA-chapter of the system (e.g. ATA25)*/

    bool Operating;/**< switch to switch the system on (1) and off (0) **/
    bool Fuel;/**< switch to define whether the system consumes fuel (1) or not (0) **/
    Vec3 refPoint;/**< [m]  Reference point of the system in global x,y,z-coordinates **/
    double Mass;/**< [kg]  system mass **/
    double scaling_factor; /**< [-]  scaling factor for the systems mass **/
    double Volume; /**< [l] system volume **/

    powerProfile heatLoad; /**  heat loads [W] */
    powerProfile bleedAir;/** bleed air [kg/s] **/
    powerProfile hydraulics;/** hydraulic power [W] **/
    powerProfile electrical;/** electric power [W] **/
    powerProfile total;/** total power [W] **/ //TODO(mazur)
    /*Hydraulic powerProfile as vector over the hydraulic circuits*/
    std::vector<powerProfile> Hydraulic;/**< [W]   **/
    std::vector<powerProfile> Electric;/**< [W]   **/

    /* Energy sources for shaft power consumption are defined by the user for certain systems.*/
    /** \brief class for the power source systems
     */
    class PowerSource {
     public:
        std::string type;/**< type of power source*/
        int sourceID;/**< source ID*/
        double operationFactor;/**< operation factor of the power source. Can be in the range [0,1] */

        PowerSource()
        :
        type(""),
        sourceID(0),
        operationFactor(0.) {}
    };

    /**
     * \brief dummy read function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    virtual void readParameters(const node& config) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(config);
    }

    /**
     * \brief dummy function for checking the user input in the configuration file for consistency
    */
    virtual void checkUserInput() {}

    /**
     * \brief dummy update function for updating the system specific values in the aircraft xml
    */
    virtual void updateXML() {}

    /** \brief dummy function to calculate the systems power. Specific functions implemented in e.g. conventionalATA25, etc.
     *
     * \param allPreviousSystemsBleed powerProfile
     * \param allPreviousSystemsHydraulic vector<powerProfile>
     * \param allPreviousSystemsElectric vector<powerProfile>
     * \param allPreviousSystemsHeat powerProfile
     * \return virtual void
     *
     */
    virtual void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                                       const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                                       const std::vector<powerProfile>& allPreviousSystemsElectric,
                                       const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    }

    /** \brief dummy function to calculate the ECS power. Specific functions implemented in conventionalATA21
     *
     * \param heatSinks powerProfile
     * \param heatECS powerProfile
     * \param heatConductor powerProfile
     * \return virtual void
     *
     */
    virtual void calculateECSPower(powerProfile heatSinks, powerProfile heatECS, powerProfile heatConductor) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(heatSinks);
    UNUSED(heatECS);
    UNUSED(heatConductor);
    }

    /** \brief dummy function to calculate the conducting power. Specific functions implemented in conducting systems (conventionalATA24, conventionalATA29, and conventionalATA36)
     *
     * \param totalBleed powerProfile
     * \param totalElectric vector<powerProfile>
     * \param totalHydraulic vector<powerProfile>
     * \return virtual void
     *
     */
    virtual void calculateConductingPower(const powerProfile& totalBleed,
                                          const std::vector<powerProfile>& totalElectric,
                                          const std::vector<powerProfile>& totalHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(totalBleed);
    UNUSED(totalElectric);
    UNUSED(totalHydraulic);
    }

    /** \brief dummy function to calculate the source power. Specific functions implemented in source systems (conventionalATA49, conventionalATA70)
     *
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return virtual void
     *
     */
    virtual void calculateSourcePower(const powerProfile& sourceBleed,
                                      const std::vector<powerProfile>& sourceElectric,
                                      const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    }

    /** \brief dummy function to calculate the sink mass. Specific functions implemented for all "sink" systems
     *
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return virtual void
     *
     */
    virtual void calculateSinkMass(const powerProfile& sourceBleed,
                                   const std::vector<powerProfile>& sourceElectric,
                                   const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    }

    /** \brief dummy function to calculate conductor mass. Specific functions implemented in conducting systems (conventionalATA24, conventionalATA29, conventionalATA36)
     *
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return virtual void
     *
     */
    virtual void calculateConductorMass(const powerProfile& sourceBleed,
                                        const std::vector<powerProfile>& sourceElectric,
                                        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    }

    /** \brief dummy function to calculate the mass of the source systems. Specific functions implemented in source systems (conventionalATA49, conventionalATA70)
     *
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return virtual void
     *
     */
    virtual void calculateSourceMass(const powerProfile& sourceBleed,
                                     const std::vector<powerProfile>& sourceElectric,
                                     const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    }

    /** \brief general constructor for all aircraft systems
     *
     * \param systemName string
     * \param systemIsOperating bool
     * \param data pointer to IOData class
     *
     */
    aircraftSystem(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~aircraftSystem();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEM_H_
