/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA32_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA32_H_

/* The following statements import necessary libraries */
#include <string>                                                           // Import c++ standard library for using string
#include <vector>                                                           // Import c++ standard library for using vector
#include <memory>                                                           // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include <svl/SVL.h>                                                        // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                              // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                                // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                               // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                            // aircraftSystem.h because that would lead to circular dependencies

/** \class conventionalATA32
 * \brief class that contains all properties of the landing gear actuation system (ATA32)
 */
class conventionalATA32: public aircraftSystem {
 public:
    int numberOfPowerSources; /**< number of power sources for ATA32 */
    std::vector<PowerSource> powerSources; /**< vector containing the power sources for ATA32 */
    double efficiencyFactor; /**< [-] landing gear chassis efficiency factor */
    double retractionTime;/**< [s] Landing gear retraction time */
    double extensionTime;/**< [s] landing gear extension time */

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                               const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                               const std::vector<powerProfile>& allPreviousSystemsElectric,
                               const powerProfile& allPreviousSystemsHeat) override;
    void calculateSinkMass(const powerProfile& sourceBleed,
                           const std::vector<powerProfile>& sourceElectric,
                           const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the power of the landing gear actuation system (ATA32)
     * \return void
     */
    void getPower();

    /** \brief calculates the mass of the landing gear actuation system (ATA32)
     * \return double: mass of the landing gear actuation system [kg]
     */
    double getMass();

    /** \brief calculates the reference point of the landing gear system (ATA32)
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether system is operating (1) or not(0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA32(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA32();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA32_H_
