/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA36_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA36_H_

/* The following statements import necessary libraries */
#include <string>                                                       // Import c++ standard library for using string
#include <vector>                                                       // Import c++ standard library for using vector
#include <memory>                                                       // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include "../aircraftSystem.h"                                          // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                            // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                           // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                        // aircraftSystem.h because that would lead to circular dependencies
#include <svl/SVL.h>                                                    // Import of header file from src folder for SVL

/** \class conventionalATA36
 * \brief class that contains all properties of the bleed air system (ATA36)
 */
class conventionalATA36: public aircraftSystem {
 public:
    double lengthDucting;/**< [m] bleed air ducting length*/
    double temperatureBleedAir;/**< [K] bleed air temperature**/
    double efficiencyFactorBleedAir;/**< [-] efficiency factor of the bleed air system */
    double specificDuctingMass;/**< [kg/m] specific ducting mass */

    //ECS values from config that are needed for calculation of bleed air
    double cabinTemperature;    /**< [K] cabin temperature**/
    double heatCapacityAir;     /**< [J/(kg*K)] heat capacity of the air**/

    //APU values from config that are needed for calculation of bleed air
    double posAPU; /**< [-] Relative position of the APU in reference to the total fuselage length*/

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;

    void calculateConductingPower(const powerProfile& totalBleed,
                                  const std::vector<powerProfile>& totalElectric,
                                  const std::vector<powerProfile>& totalHydraulic) override;
    void calculateConductorMass(const powerProfile& sourceBleed,
                                const std::vector<powerProfile>& sourceElectric,
                                const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief  calculates the power demand of bleed air system (ATA36)
     * \return void
     */
    void getPower(powerProfile theBleed);

    /** \brief calculates the mass of the bleed air system (ATA36)
     * \return double: mass of the bleed air system [kg]
     */
    double getMass();

    /** \brief calculates the reference point of the bleed air system (ATA36)
     * \return Vec3: reference point of the bleed air system in global x, y , z-coordinates [m]
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether system is operating (1) or not (0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA36(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA36();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA36_H_
