/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_GEOMETRICALFUNCTIONS_H_
#define SYSTEMS_DESIGN_SRC_GEOMETRICALFUNCTIONS_H_

#include <vector>
#include <string>
#include <unordered_map>

#include <svl/SVL.h>
#include "systemsIOData.h"

class nacelle;

/** \brief contains geometrical functions for the aircraft
 */
class geometricalFunctions {
 public:
    /** \brief calculates the conductor length from wing-mounted engines to the wing root.
     *
     * \param theWing liftingSurface
     * \param theNacelles vector<nacelle>
     * \return double: [m] conductor length
     *
     */
    double getConductorLengthWingEngines(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing,
              const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles) const;

    /** \brief calculates the conductor length along a airfoil surface trailing edge.
     *
     * \param theWing const geom2::MultisectionSurface<geom2::AirfoilSection>
     * \return double: [m] conductor length
     *
     */
    double getConductorLengthSurfaceTrailingEdge(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing);

    /** \brief calculates the conductor length from fuselage mounted engines to fuselage midline.
     *
     * \param theFuselage fuselage
     * \param theNacelles vector<nacelle>
     * \return double: [m] conductor length
     *
     */
    double getConductorLengthFuselageEngines(const geom2::MultisectionSurface<geom2::PolygonSection>& theFuselage,
              const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles) const;

    /** \brief calculates the conductor length from fuselage-mounted engines to a system position.
     *
     * \param Pos1 double: [m] first x position in global coordinates
     * \param Pos2 double: [m] sedond x position in global coordinates
     * \param theNacelles vector<nacelle>
     * \return double: [m] conductor length
     *
     */
    double getConductorLengthFuselageEnginesToSystem(double Pos1, double Pos2,
              const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles);

    /** \brief Function to collect the relative start and end position for de-icing.
     * \param theWing the main wing
     * \param wingRelSpanAntiIcingStart_str string holding either DEFAULT or relative start position
     * \return unordered_map holding relative spanwise start and end positions
     */
    std::unordered_map<std::string, double> getRelSpanAntiIcing(const systemsIOData::Wing& theWing, const std::string& wingRelSpanAntiIcingStart_str);

    /** \brief Constructor
     *
     *
     */
    geometricalFunctions();
    virtual ~geometricalFunctions();

 private:
    /** \brief calculates the wing coordinates of the wing-mounted nacelle
     *
     * \param yPos double: [m]
     * \param theWing geom2::MultisectionSurface<geom2::AirfoilSection>
     * \return Vec3: [m] global coordinates
     *
     */
    Vec3 getWingCoodinatesOfNacelle(double yPos, const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing) const;

    /** \brief sorts coordinates in y-direction
     *
     * \param Coordinates vector<Vec3>&: [m] global coordinates in x,y,z
     * \return void
     *
     */
    void sortCoordinates(std::vector<Vec3> *Coordinates) const;

    /** \brief calculates the partial length of a conductor through the wing
     *
     * \param Pos1 Vec3: [m] position 1 of the conductor in global x,y,z coordinates
     * \param Pos2 Vec3: [m] position 2 of the conductor in global x,y,z coordinates
     * \param theWing geom2::MultisectionSurface<geom2::AirfoilSection>
     * \return double: [m] condctor length
     *
     */
    double getPartialLengthWingConductor(Vec3 Pos1, Vec3 Pos2, const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing) const;

    /** \brief calculates the relative start position for de-icing. If start position is DEFAULT, start position of first outboard LE device is identified.
     * \param theWing the main wing
     * \param wingRelSpanAntiIcingStart_str string holding either DEFAULT or relative start position
     * \return double: relative spanwise start position [-]
     */
    double getRelSpanAntiIcingStart(const systemsIOData::Wing& theWing, const std::string& wingRelSpanAntiIcingStart_str) const;

    /** \brief calculates the relative end position for de-icing.
     * \param theWing the main wing
     * \return double relative spanwise end position [-]
     */
    double getRelSpanAntiIcingEnd(const systemsIOData::Wing& theWing) const;
};

#endif // SYSTEMS_DESIGN_SRC_GEOMETRICALFUNCTIONS_H_
