# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""File providing functions to read layer text from aircraft XML file."""
# Import standard libraries.
import sys


def read_energy_carrier(root_of_aircraft_exchange_tree, runtime_output):
    """Read energy carrier from aircraft exchange file.

    This function extracts information about the energy carrier used in an aircraft from the provided aircraft exchange
    file. It specifically looks for 'energy_carriers' nodes and their corresponding 'energy_carrier' sub-nodes.

    :param ElementTree root_of_aircraft_exchange_tree: Root of aircraft exchange XML
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :raises ValueError: Raised if 'energy_carriers' or 'energy_carrier' node does not exist
    :return string energy_carrier: Energy carrier string
    """

    # Initialize empty list.
    energy_carrier_list = []

    # Attempt to extract information on used energy carrier from aircraft exchange file.
    try:
        # Find all 'energy_carriers' nodes in aircraft exchange file.
        energy_carrier_node_list = root_of_aircraft_exchange_tree.findall('.//energy_carrier')
        # Check, if 'energy_carrier' nodes exist.
        if not energy_carrier_node_list:
            # Raise error, if no 'energy_carrier' node exists.
            raise ValueError('No "energy_carrier" nodes found in the aircraft exchange file. Program aborted.')
        # Iterate over 'energy_carrier_node_list' and append values of energy carrier sub-nodes.
        for carrier_node in energy_carrier_node_list:
            energy_carrier_node = carrier_node.find('.//value')
            if energy_carrier_node is not None:
                energy_carrier_list.append(energy_carrier_node.text)
            else:
                raise ValueError('No energy carrier nodes found in the aircraft exchange file. Program aborted.')

        # If 'energy_carrier_list' is not empty, compare all entries.
        if energy_carrier_list is not None:
            # If all entries are the same, set 'energy_carrier' to first list entry.
            if all(element == energy_carrier_list[0] for element in energy_carrier_list):
                energy_carrier = energy_carrier_list[0]
            # If list entries differ, set 'energy_carrier' to 'hybrid'.
            else:
                energy_carrier = 'hybrid'
        # Raise error, if 'energy_carrier_list' is empty.
        else:
            raise ValueError('No energy carrier node found in "energy_carriers" nodes. Program aborted.')

    # Exception handling for ValueError.
    except ValueError as e:
        runtime_output.critical('Error: ' + str(e))
        sys.exit(1)

    return energy_carrier


def read_tank_configuration(root_of_aircraft_exchange_tree, runtime_output):
    """Read tank configuration.

    Read tank configuration information from aircraft exchange file.

    :param ElementTree root_of_aircraft_exchange_tree: Root of aircraft exchange XML
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return str tank_configuration: Information on tank configuration
    """

    # Initialize empty list.
    tank_configuration_dict = {}

    # Attempt to extract information on tank configuration from aircraft exchange file.
    try:
        # Find all 'tank' nodes in aircraft exchange file.
        tank_node_list = root_of_aircraft_exchange_tree.findall('.//tank_configuration/tank')
        # Check, if 'tank' nodes exist.
        if not tank_node_list:
            # Raise error, if no 'tank' node exists.
            raise ValueError('No "tank" nodes found in the aircraft exchange file. Program aborted.')
        # Iterate over 'tank_node_list' and append values of 'energy_carrier_ID' sub-nodes.
        for tank_node in tank_node_list:
            # Check if all necessary parameters exist.
            tank_id_exists = tank_node.get('ID')
            energy_carrier_id_exists = tank_node.find('.//energy_carrier_ID/value')
            location_exists = tank_node.find('.//location/value')
            position_exists = tank_node.find('.//position/value')
            if (tank_id_exists is None or energy_carrier_id_exists is None
                or location_exists is None or position_exists is None):
                raise ValueError('Incomplete tank information found in the aircraft exchange file. Program aborted.')
            # Extract necessary parameters.
            tank_id = tank_node.get('ID')
            location = tank_node.find('.//location/value').text
            position = tank_node.find('.//position/value').text
            # Get according energy carrier.
            energy_carrier_id = tank_node.find('.//energy_carrier_ID/value').text
            energy_carrier = (
                root_of_aircraft_exchange_tree.find('.//energy_carrier[@ID="' + energy_carrier_id + '"]/value').text)
            tank_configuration_dict['tank_' + tank_id] = {'energy_carrier': energy_carrier,
                                                          'location': location,
                                                          'position': position}

    # Exception handling for ValueError.
    except ValueError as e:
        runtime_output.critical('Error: ' + str(e))
        sys.exit(1)

    return tank_configuration_dict
