/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "sweepAngleMethods.h"

#include <aircraftGeometry2/processing/transform.h>
#include <atmosphere/atmosphere.h>
#include <unitConversion/constants.h>

#include <functional>
#include <map>

#include "lib/general_methods/bisectionAlgorithm.h"
double user_defined::sweep_angle_at_quarter_chord(const double user_sweep_angle_at_quarter_chord) {
  return user_sweep_angle_at_quarter_chord;
}

double drag_divergence::sweep_angle_at_quarter_chord(const double mach_design, const double design_altitude,
                                                     const double delta_drag_divergence,
                                                     const double technology_factor_drag_divergence,
                                                     const double max_thickness_to_chord_ratio,
                                                     const double wing_loading) {
  constexpr double kappa_air = 1.4;
  constexpr double climb_mass_loss = 0.02;
  atmosphere isa;

  double pressure_ica_design_altitude = isa.getPressureISA(design_altitude);
  const double mach_drag_divergence = mach_design + delta_drag_divergence;

  auto lift_coefficient_by_korn_adapted = [=](const double sweep) {
    return 10. * technology_factor_drag_divergence * pow(cos(sweep), 2.) -
           10. * max_thickness_to_chord_ratio * cos(sweep) - 10. * mach_drag_divergence * pow(cos(sweep), 3.0);
  };
  auto lift_coefficient_by_mach = [=](const double mach) {
    return 2. * G_FORCE * wing_loading * (1 - climb_mass_loss) /
           (kappa_air * pressure_ica_design_altitude * pow(mach, 2.));
  };
  const double lift_coefficient_mach_drag_divergence = lift_coefficient_by_mach(mach_drag_divergence);

  auto sweep_searcher = [=](const double sweep) {
    return lift_coefficient_by_korn_adapted(sweep) - lift_coefficient_mach_drag_divergence;
  };
  const double quarter_chord_sweep = bisection_algorithm(sweep_searcher, 0., 30. * geom2::detail::to_radians);

  return quarter_chord_sweep;
}

double get_wing_quarter_chord_sweep(modes quarter_chord_sweep_mode, const double user_quarter_chord_sweep,
                                    const double mach_ica, const double wing_loading, const double design_altitude,
                                    const double max_thickness_to_chord_ratio, const double delta_drag_divergence,
                                    const double technology_factor_drag_divergence) {
  std::map<modes, std::tuple<std::string, std::function<double(void)>>> quarter_chord_sweep_selection = {
      {"mode_0",
       {"user_defined", [=]() { return user_defined::sweep_angle_at_quarter_chord(user_quarter_chord_sweep); }}},
      {"mode_1", {"drag_divergence", [=]() {
                    return drag_divergence::sweep_angle_at_quarter_chord(
                        mach_ica, design_altitude, delta_drag_divergence, technology_factor_drag_divergence,
                        max_thickness_to_chord_ratio, wing_loading);
                  }}}};
  const auto [method, quarter_chord_sweep_method] = quarter_chord_sweep_selection[quarter_chord_sweep_mode];
  double quarter_chord_sweep = quarter_chord_sweep_method();
  std::cout << "Quarter chord sweep [method]: " << method << "\n";
  std::cout << "Quarter chord sweep  [value]: " << quarter_chord_sweep * geom2::detail::to_degrees << " [deg]\n";
  return quarter_chord_sweep;
}